# Prototype Lua Object-Oriented Program System

**PLoop** is an object-oriented program system for the Lua.

It's built on a prototype system which is used to easy the using of the Lua meta-table system.


## Prototypes

Based on the prototype system, the **PLoop** provided several prototypes:

* The **environment** prototype provide the common environmental controls for code management and type building.

* The **attribute** prototype provide the functional description labels for the codes like function, class, object and etc.

* The **namespace** prototype provide a namespace system to manage and provide quick acessing of types.

* The **enum** prototype provide the enumeration validation data types.

* The **struct** prototype provide the structure validation data types.

* The **interface** prototype provided the interface system, interfaces are abstracts of functions, also are the multiple inheritance system provided by the **PLoop**.

* The **class** prototype provided the class system, the classes are abstracts of the objects, it provide method, meta-method, features(see below) to the objects.

* The **property** prototype provided the property features to the class objects.

* The **event** prototype provided the event features to the class objects.


## PLoop

When load the **PLoop**, we can use the variable *PLoop* to access features defined in it or use it to process the codes. The **PLoop** won't saved keywords like *class*, *namespace* and etc into the `_G` if they already existed, but'll always save the *PLoop* to the `_G`, so we can use it as the only acess point of the **PLoop** system.

It's also returned by the `require "PLoop"`. Here is a simple example:

```lua
require "PLoop" (function(_ENV)
	print(XList(100):Sum())  -- Sum 1 to 100
end)
```

The PLoop is the root namespace of all others. It's an anonymous namespace, we can only access it by the *PLoop* variable. Normally we only need use it when use **PLoop**'s features in the `_G`.

The above example can be converted to

```lua
require "PLoop"

-- Use PLoop to access its sub namespaces
print(PLoop.System.Collections.XList(100):Sum())
```


## Spell Error Checks And More

There are a lots of troubles in the Lua debugging, if the lua error can be triggered, it's still easy to fix it, but for codes like `if a == ture then`, *ture* is a non-existent variable, Lua treate it as nil so the checking will still working, but the result can't be right.

We'll see how to solve it in the **PLoop**.


### Read un-existed global variables

Before rquire the **PLoop**, we can create a **PLOOP_PLATFORM_SETTINGS** table to toggle the **PLoop**'s system settings:

```lua
PLOOP_PLATFORM_SETTINGS = { ENV_ALLOW_GLOBAL_VAR_BE_NIL = false }

require "PLoop"

PLoop(function(_ENV)
	local a = ture  -- Error: The global variable "ture" can't be nil.

	if a then
		print("ok")
	end
end)
```

Turn off the **ENV_ALLOW_GLOBAL_VAR_BE_NIL** will apply a strict mode for all **PLoop** private environment, so no nil variables can be accessed, so you can locate those errors.


### Write to illegal global variables

If we missing the `local`, we may create unwanted global variables. But the system can't diff the wanted and unwanted global variable, we can add filter in the platform settings to do the job, so we can remove the filter when we don't need it:

```lua
PLOOP_PLATFORM_SETTINGS = {
	GLOBAL_VARIABLE_FILTER = function(key, value)
		-- Don't allow the lowercase key with non-function value
		if type(key) == "string" and key:match("^%l") and type(value) ~= "function" then
			return true
		end
	end,
}

require "PLoop"

PLoop(function(_ENV)
	Test = 1

	class "A" (function(_ENV)
		function Test(self)
			ch = 2 -- error: There is an illegal assignment for "ch"
		end
	end)

	A():Test()
end)
```

If the filter return true, the assignment will trigger an error, so the code'll be stopped, if we only need a warning, we can add a setting like:

```lua
PLOOP_PLATFORM_SETTINGS = {
	GLOBAL_VARIABLE_FILTER = function(key, value)
		-- Don't allow the lowercase key with non-function value
		if type(key) == "string" and key:match("^%l") and type(value) ~= "function" then
			return true
		end
	end,
	GLOBAL_VARIABLE_FILTER_USE_WARN = true,
}

require "PLoop"

PLoop(function(_ENV)
	Test = 1

	class "A" (function(_ENV)
		function Test(self)
			ch = 2 -- [PLoop: Warn]There is an illegal assignment for "ch"@path_to_file\file.lua:18
		end
	end)

	A():Test()
end)
```

You also can use the filter as a record, with another setting, the call line'll be passed in as the 3rd argument:

```lua
PLOOP_PLATFORM_SETTINGS = {
	GLOBAL_VARIABLE_FILTER = function(key, value, path)
		print("Assign '" .. key .. "'" .. path )
	end,
	GLOBAL_VARIABLE_FILTER_GET_CALLLINE = true,
}

require "PLoop"

PLoop(function(_ENV)
	Test = 1  -- Assign 'Test'@path_to_file\file.lua:11

	class "A" (function(_ENV)
		function Test(self)
			ch = 2 -- Assign 'ch'@path_to_file\file.lua:15
		end
	end)

	A():Test()
end)
```

To use the get call line, the `debug.getinfo` must existed.


### Access un-existed object fields

We also can block the accessing of un-existed object fields:

```lua
PLOOP_PLATFORM_SETTINGS = { OBJECT_NO_RAWSEST = true, OBJECT_NO_NIL_ACCESS = true }

require "PLoop"

PLoop(function(_ENV)
	-- Define a class with Name and Age property
	class "Person" (function(_ENV)
		property "Name" { type = String }
		property "Age"  { type = Number }
	end)

	o = Person()

	o.Name = "King" -- Ok

	o.name = "Ann"  -- Error: The object can't accept field that named "name"

	print(o.name)   -- Error: The object don't have any field that named "name"
end)
```

This three settings will help authors to avoid many spell errors during the development. You shouldn't use those settings when you release the project since the access speeding should be slightly increased.


## Type Validation

**PLoop** make the Lua as a strong type language, there are many type validation features to stop the errors spread to far so too hard to be tracked.

The function validation is always a complex part, we need to do many checks before the function's main logic for the arguments so we can tell the caller where and what is failed. And when the project is released, those check should be removed since we already test them.

Within the **PLoop**, it'll be a small problem:

```lua
require "PLoop"

PLoop(function(_ENV)
	__Arguments__{ String, Number }
	function SetInfo(name, age)
	end

	-- Error: Usage: SetInfo(System.String, System.Number) - the 2nd argument must be number, got boolean
	SetInfo("Ann", true)
end)
```

The `__Arguments__` is an attribute class defined in the **System**, it associated the argument name, type, default value and etc to the argument, also wrap those functions with the argument validation.

The **String** and **Number** are struct types used to validate values, we'll see them at the introduction of struct.

If we need to release the project, there is also no need to remove those `__Arguments__`, you can change the platform setting( not all type validation would be removed, but just leave them to the system):

```lua
PLOOP_PLATFORM_SETTINGS = { TYPE_VALIDATION_DISABLED = true }

require "PLoop"

PLoop(function(_ENV)
	__Arguments__{ String, Number }
	function SetInfo(name, age)
	end

	-- No error now
	SetInfo("Ann", true)
end)
```

There are three type arguments:

* Number     - The required argument, means the value can't be nil and need match the target type
* Number/nil - The optional argument, means the value can be nil, or must match the target type
* Number/1   - The optional argument, means the value can be nil, the default value (1) will be used, if not nil, the value must match the target type
* Number * 0 - The varargs argument, means the last variables must match the target type, and there could have no variables
* Number * 3 - The varargs argument, means the last variables must match the target type, and there must have at least (3) variables

We'll see more usages in other topics.


## PLOOP_PLATFORM_SETTINGS

Before loading the **PLoop** lib, we can create a table named **PLOOP_PLATFORM_SETTINGS**, the table contains several settings can be modified based on the target platform and frameworks. Its fields are all optional.

We'll learn more in the next chapters.

* ATTR_USE_WARN_INSTEAD_ERROR           - Whether the attribute system use warning instead of error for invalid attribute target type. Default false

* ENV_ALLOW_GLOBAL_VAR_BE_NIL           - Whether the environmet allow global variable be nil, if false, things like ture(spell error) could trigger error. Default true

* GLOBAL_VARIABLE_FILTER                - The filter function used to validation the new declared global variables with key and value, if the function return non-false value, that means the assignment is not legal(if we write the code `idx = 1` in a function, we normally missed the `local`), so the assignment should be canceled or warned, you also can use the filter to record the global assignments

* GLOBAL_VARIABLE_FILTER_USE_WARN       - Whether use warning instead of the error, if use warning, the assignment will still be processed

* GLOBAL_VARIABLE_FILTER_GET_CALLLINE   - Whether pass the callline of the global assignment to the filter

* TYPE_DEFINITION_WITH_OLD_STYLE        - Whether allow old style of type definitions like : class "A"  endclass "A". Default false

* TYPE_VALIDATION_DISABLED              - Whether the type validation should be disabled. The value should be false during development, toggling it to true will make the system ignore the value valiation in several conditions for speed. Default false

* NAMESPACE_NIL_VALUE_ACCESSIBLE 		- Whether allow accessing non-existent value from namespace, if true access features like 'System.Nonexist' will return nil value, otherwise an error will be raised. Default true

* CLASS_NO_MULTI_VERSION_CLASS          - Whether all old objects keep using new features when their classes or extend interfaces are re-defined. Default false

* CLASS_NO_SUPER_OBJECT_STYLE           - Whether all interfaces & classes only use the classic format `super.Method(obj, ...)` to call super's features, don't use newstyle like : `super[obj].Name = "Ann"`, `super[obj].OnNameChanged = super[obj].OnNameChanged + print`, `super[obj]:Greet("King")`. Default false

* INTERFACE_ALL_ANONYMOUS_CLASS         - Whether all interfaces has anonymous class, so it can be used to generate object. Default false

* OBJECT_NO_RAWSEST                     - Whether all class objects can't save value to fields directly, So only init fields, properties, events can be set during runtime. Default false

* OBJECT_NO_NIL_ACCESS                  - Whether all class objects can't fetch nil value from it, combine it with **OBJ_NO_RAWSEST** will force a strict mode for development. Default false

* OBJECT_DEBUG_SOURCE                   - Whether save the creation places (source and line) for all objects. Default false

* CORE_LOG_LEVEL                        - The Log level used in the Prototype core part., 1(Trace), 2(Debug), 3(Info), 4(Warn), 5(Error), 6(Fatal). Default 3(Info)

* CORE_LOG_HANDLER                      - The core log handler works like : function CORE_LOG_HANDLER(message, loglevel) end.  Default print

* MULTI_OS_THREAD                       - Whether the system is used in a platform where multi os threads share one lua-state, so the access conflict can't be ignore. Default false

* MULTI_OS_THREAD_LUA_LOCK_APPLIED      - Whether the system is used in a platform where multi os threads share one lua-state, and the lua_lock and lua_unlock apis are applied, so PLoop don't need to care about the thread conflict. Default false

* MULTI_OS_THREAD_ENV_AUTO_CACHE_WARN   - Whether the system send warning messages when the system is used in a platform where multi os threads share one lua-state, and global variables are saved not to the environment but an inner cache, it'd solve the thread conflict, but the environment need fetch them by `__index` meta-call, so it's better to declare local variables to hold them for best access speed. Default true

* UNSAFE_MODE                           - Whether the system use tables for the types of namespace, class and others, and save the type's meta data in themselves. Normally it's not recommended. Default false

* EXCEPTION_SAVE_STACK_DATA             - Whether try to save the stack data into the exception object, so we can have more details about the exception. Default true

* EXCEPTION_SAVE_VARIABLES              - Whether alwasy try to save local variables and upvalues for exception

* THREAD_POOL_MAX_SIZE                  - The max pool size of the thread pool. Default 40

* USE_DISPOSE_AS_META_GC 				-- Use the Dispose as the `__gc`, only works for Lua 5.3 and above. Default true


### Recommend Settings - Development

With several settings, the **PLoop** could help the developers to quickly locate spell and type errors.

```lua
PLOOP_PLATFORM_SETTINGS 			= {
    OBJECT_NO_RAWSEST 				= true,
    OBJECT_NO_NIL_ACCESS 			= true,
    TYPE_VALIDATION_DISABLED 		= false,
    ENV_ALLOW_GLOBAL_VAR_BE_NIL 	= false,
    NAMESPACE_NIL_VALUE_ACCESSIBLE 	= false,
}

require "PLoop" (function(_ENV)
	class "Person" { Name = String, Age = Number }

	ann = Person{ Name = "Ann", Age = 12 }

	ann.name = "Lucy" 		-- Forbidden by OBJECT_NO_NIL_ACCESS for non-existed property/event 'name'
	ann.Age = ann.age + 1  	-- Forbidden by OBJECT_NO_RAWSEST for non-existed property/event 'age'

	__Arguments__{ Number }
	function test(id) end

	test("hi") 				-- Forbidden by TYPE_VALIDATION_DISABLED(false) for not pass the argument check

	print(ture) 			-- Forbidden by ENV_ALLOW_GLOBAL_VAR_BE_NIL(false) for non-existed global variable

	print(System.Xyz) 		-- Forbidden by NAMESPACE_NIL_VALUE_ACCESSIBLE(false) for non-existed namespace
end)
```

### Recommend Settings - Release

Although those checking features are very useful during the development, but they also give the system a tiny performance loss. To gain more performance, it's better to turn them off for the release build:

```lua
PLOOP_PLATFORM_SETTINGS 			= {
    OBJECT_NO_RAWSEST 				= false,
    OBJECT_NO_NIL_ACCESS 			= false,
    TYPE_VALIDATION_DISABLED 		= true,
    ENV_ALLOW_GLOBAL_VAR_BE_NIL 	= true,
    NAMESPACE_NIL_VALUE_ACCESSIBLE 	= true,
}
```