# environment

The **environment** is a prototype used to provide the environment controls.

The environments are designed to be private and standalone for codes(module) or type building(class and etc). Within them we can use features like keyword accessing, namespace management, get/set management and etc.

The **environment** prototype can't be accessed directly, but it's used everywhere in the **PLoop**.

**You can skip this chapter if you don't have problem with PLoop's code style, the System.Module part also will be enough if you need isolated code environment.**


## Table of Contents

* [Code Isolated](#code-isolated)
* [The global variable access](#the-global-variable-access)
* [Auto-cache](#auto-cache)
* [Attribute for global functions](#attribute-for-global-functions)
* [Type builder & Keyword accessing](#type-builder--keyword-accessing)
* [System.Environment](#systemenvironment)
	* [Apply](#apply)
	* [BackupKeywordAccess](#backupkeywordaccess)
	* [ExportVariables](#exportvariables)
	* [GetNamespace](#getnamespace)
	* [GetParent](#getparent)
	* [GetValue](#getvalue)
	* [GetKeywordVisitor](#getkeywordvisitor)
	* [ImportNamespace](#importnamespace)
	* [Initialize](#initialize)
	* [RegisterGlobalNamespace](#registerglobalnamespace)
	* [RegisterContextKeyword](#registercontextkeyword)
	* [RegisterRuntimeContextKeyword](#registerruntimecontextkeyword)
	* [RegisterGlobalKeyword](#registerglobalkeyword)
	* [RegisterRuntimeKeyword](#registerruntimekeyword)
	* [RestoreKeywordAccess](#restorekeywordaccess)
	* [SaveValue](#savevalue)
	* [SetDefinitionMode](#setdefinitionmode)
	* [SetNamespace](#setnamespace)
	* [SetParent](#setparent)
* [System.Module](#systemmodule)
	* [Version Control](#version_control)
	* [Usable features](#usable_features)
	* [Create a new module system](#create_a_new_module_system)


## Code Isolated

The codes are isolated by those environments, so we can enjoy the global variables's benefit without
polluting the `_G` and other environments.

```lua
require "PLoop"

PLoop(function(_ENV)
	function Test() end
end)

print(Test) -- nil

PLoop(function(_ENV)
	print(Test)  -- nil
end)
```

Use the **PLoop** to call a function is the basic way to create the object of the environment prototype and use it as the function's environment, that's also why we need the `_ENV` to be the first parameter of the funciton(it make sure the same code can be ran at both Lua 5.1 and above versions).

Since the codes in the function is processed in the private environment, it won't effect others.

We'll see more about the codes isolation in the **System.Module** part.


## The global variable access

Coding in the private environment is almost the same like we do in the `_G`, the different is the private environment is empty when created, it has no global variables like *print*, *math* and etc. But we still can use them.

Within the private environment, we can access any global variables defined in the `_G`, also we can access features provided by the **PLoop** through the namespace system.

When a global variable isn't existed in the private environment and the codes accessed it, the private environment should check it with orders:

1. Try fetch types with the name from the namespace system, we'll see more in the namespace chapter. If nil go to step 2, else return the value.

2. Try fetch value from the base environment, the private environment always have a base environment, default is the `_G`. The system also will check the base environment's base environment until it's the `_G`. If nil go to step 3, else return the value.

3. If the **PLOOP_PLATFORM_SETTINGS.ENV_ALLOW_GLOBAL_VAR_BE_NIL** is false, raise an error since it's more likely we have a spell error there, else return nil as the value.

```lua
PLOOP_PLATFORM_SETTINGS = { ENV_ALLOW_GLOBAL_VAR_BE_NIL = false }

require "PLoop"

PLoop(function(_ENV)
	print(List) -- fetch "print" from the _G, fetch List from namespaces

	local a = ture -- Error: The global variable "ture" can't be nil.
end)
```


## Auto-cache

To improve the performance, when the private environment access a global variable not existed in itself, it'll try to cache the value in it during runtime, but won't do the auto-cache job during the definition phase.

```lua
require "PLoop"

PLoop (function(_ENV)
	-- System.Collections.List  nil
	print(List, rawget(_ENV, "List"))

	_G.Dojob = function()
		print(List, rawget(_ENV, "List"))
	end
end)

-- System.Collections.List  System.Collections.List
Dojob()
```

When the function called by the **PLoop** processing, the codes are running in definition phase, and those codes only will be executed for once, there is no need to cache those global variables since they only will be accessed for one time.

When we call the `Dojob()`, the codes is processed in runtime phase, we may call the function again and again, so it's time to save those global variables in the private environment.

Besides the auto-cache mechanism, there is no rule to block you using the code like  `local print = print`.


## Attribute for global functions

The **PLoop** has provide an attribute system used to modify, apply changes or attach data to code features like class, object, function and etc.

Any global functions defined in the private environment can have attributes, so we can easily modify those functions or register them for other usages.

```lua
require "PLoop" (function(_ENV)
	__Iterator__()
	function factorial(max)
		local v = 1
		for i = 1, max do
			v = v * i
			coroutine.yield(i, v)
		end
	end

	-- 1    1
	-- 2    2
	-- 3    6
	for i, f in factorial(3) do
		print(i, f)
	end
end)
```

The `__Iterator__` attribute is used to modify the target function so it can be used as an iterator, we can use `yield` to return values from the function for multi times.

We'll see more attributes in other examples, and the collection system for more iterator examples.


## Type builder & Keyword accessing

**PLoop** use the private environment to building types. Here is an example for class:

```lua
require "PLoop"

class "A" (function(_ENV)
	property "Name" { type = String }
end)
```

The **property** is a keyword only usable in the interface or class definitions, it's accessed through the `_ENV`. The keyword also will know who is its caller through the `_ENV` and add the *Name* property to the *A* class.


## System.Environment

**System.Environment** is the proxy created from the **environment** prototype. It contains all features of the **environment** system.

### Apply

Apply the environment to the function or stack, for Lua 5.2 and above versions, may not be able apply the environment to the target stack.

* Format:
	* (env, func)
	* (env[, stack])

* Params:
	* env           - the environment
	* func          - function, the target function
	* stack         - number, the target stack level, default 1


### BackupKeywordAccess

Back up the accessed keyword and visitor, **should only be used by the system**.

* Return
	* data          - the keyword access information


### ExportVariables

Export variables by name or a list of names, those variables are fetched from the namespaces or base environment.

* Format: (env, variables[, stack])
* Params:
	* env           - the environment
	* variables     - the variable name or a list of the variable name or a table contains the key-value pairs
	* stack         - the stack level

In the multi-thread platform, to avoid the re-hash conflict, the auto-cache mechanism still works but with a price, so it's better to save them to the environment directly, we don't need use this function directly:

```lua
require "PLoop" (function(_ENV)
	export "print"
	export { "math", "pairs", List }
	export { ipairs = ipairs }
	export {
		"select",
		abs = math.abs,
		Dictionary,
	}

	print(rawget(_ENV, "select"))
end)
```

Since the types defined in the **PLoop** have their names contains in themselves, we can use them directly.


### GetNamespace

Get the namespace from the environment, see **SetNamespace** for details.

* Params:
	* env           - the environment
* Return:
	* ns            - the environment's namespace


### GetParent

Get the parent environment from the environment, See **SetParent** for details.

* Params:
	* env           - the environment
* Return:
	* parent        - the parent environment


### GetValue

Get the value from the environment based on its namespace and parent settings(normally be used in `__newindex` for environment), the keywords also must be fetched through it.

* Format: (env, name[, noautocache][, stack])
* Params:
	* env           - the environment
	* name          - the query name
	* noautocache   - true if don't save the value to the environment, the keyword won't be saved
	* stack         - the stack level
* Return:
	* value         - the query value

This is normally used as the environment's `__index` meta-method. If you don't nee to create your own type builders, there is no need to use it.


### GetKeywordVisitor

Get the environment that visit the given keyword. The visitor use **Environment.GetValue** to access the keywords, so the system know where the keyword is called, this method is normally called by the keywords.

* Params :
	* keyword       - the keyword
* Return
	* visitor       - the keyword visitor

If you don't need to create your own keywords, there is no need to use it.


### ImportNamespace

Import namespace to environment, an environment can import unlimited namespaces, the **Environment.GetValue** method will check those imported namespaces for types.

* Format: (env, ns[, stack])
* Params:
	* env           - the environment
	* ns            - the namespace or its path

```lua
require "PLoop" (function(_ENV)
	import "System.Serialization"
	-- or
	import (System.Serialization)
end)
```

### Initialize

Initialize the environment, must be called when a new environment is created, normally already done by the system.

* Params:
	* env           - the environment


### RegisterGlobalNamespace

Register a namespace as global namespace, so it can be accessed by any environment.

* Params:
	* ns            - the namespace or its path

There are several global namespaces provided by the **PLoop** so we can access type like **List** from **System.Collections**, **Number** from **System** directly.

* System
* System.Collections
* System.Threading


### RegisterContextKeyword

Register a context keyword, like **property** must be used in the definition of a class or interface. This keyword should only be used during definition phase.

* Format:
	* (contexttype, [key, ]keyword)
	* (contexttype, keywords)
* Parmas:
	* contexttype   - the context environment's type
	* key           - the keyword's name, it'd be applied if the keyword is a function
	* keyword       - the keyword entity
	* keywords      - a collection of the keywords like : { import = import , class, struct }


### RegisterRuntimeContextKeyword

Register a context keyword, like **this** used for contructors. This keyword should only be used during runtime.

* Format:
	* (contexttype, [key, ]keyword)
	* (contexttype, keywords)
* Parmas:
	* contexttype   - the context environment's type
	* key           - the keyword's name, it'd be applied if the keyword is a function
	* keyword       - the keyword entity
	* keywords      - a collection of the keywords like : { import = import , class, struct }


### RegisterGlobalKeyword

Register a global keyword, like **class** can be used anywhere. This keyword should only be used during definition phase.

* Format:
	* ([key, ]keyword)
	* (keywords)
* Parmas:
	* key           - the keyword's name, it'd be applied if the keyword is a function
	* keyword       - the keyword entity
	* keywords      - a collection of the keywords like : { import = import , class, struct }


### RegisterRuntimeKeyword

Register a global keyword, like **class** can be used anywhere. This keyword should only be used during runtime.

* Format:
	* ([key, ]keyword)
	* (keywords)
* Parmas:
	* key           - the keyword's name, it'd be applied if the keyword is a function
	* keyword       - the keyword entity
	* keywords      - a collection of the keywords like : { import = import , class, struct }


### RestoreKeywordAccess

Restore the accessed keyword and visitor, **should only be used by the system**.

* Params:
	* data          - the data returned by **BackupKeywordAccess**


### SaveValue

Save the value to the environment

* Format: (env, name, value[, stack])
* Params:
	* env           - the environment
	* name          - the value's key
	* value         - the target value
	* stack         - the stack level

This is normally used as environment's `__newindex` meta-method.


### SetDefinitionMode

Turn on/off the definition mode for an environment, the value won't be auto-cached to the environment in definition mode

* Params:
	* env           - the environment
	* mode          - boolean


### SetNamespace

Set the namespace to the environment, an environment can have and only have one namespace as its own, all types defined in it will saved to the namespace.

* Format: (env, ns[, stack])
* Params:
	* env           - the environment
	* ns            - the namespace
	* stack         - the stack level

We don't use the method directly, there is a global keyword to do the job:

```lua
require "PLoop" (function(_ENV)
	namespace "System.Serialization" -- create and set the namespace
	-- or
	namespace (System.Serialization) -- set the namespace
end)
```


### SetParent

Set the parent environment to the environment, the parent is the base environment of the target environment, so the target environment can fetch global values from its parent.

* Format: (env, base[, stack])
* Params:
	* env           - the environment
	* base          - the parent environment
	* stack         - the stack level


## System.Module

In the previous examples, all private environments are stand-alone, they share the global variables from `_G`, but won't share any global variables defined in themselves.

A class **System.Module** is defined to provide a controllable module system, which share the global variables with a tree structure.

```lua
require "PLoop"

_ENV = Module "Test" "1.0.0"  -- Use PLoop.System.Module instead of Module if you have Module from other libs

function Test()
	print("This is defined  in the Test module")
end
```

```lua
_ENV = Module "Test.A" "1.0.0"

function TestA() end

print(Test)  -- func:xxxxxxx
```

```lua
_ENV = Module "Test.B" "1.0.0"

print(Test)  -- func:xxxxxx
print(TestA) -- nil
```

```lua
_ENV = Module "Test.A.B.C.D.E.F" "1.0.0"

print(Test)  -- func:xxxxxx
print(TestA) -- func:xxxxxx
```

The **System.Module** is a class, we can create its objects like `Module "Test.A.B.C.D"`, and then use `_ENV = module "[version]"` to change the file's current environment to the module, so all the codes is processed in the modules.

We also can use function as the code container like :

```lua
Module "Test.A" (function(_ENV)
	print(Test)  -- func:xxxxxx
	print(TestA) -- func:xxxxxx
end)
```

For the given module name, it'd be split by the dot, each will be the name of a module who is a child module of the previous one, so the `Module "Test"` is the root module, and `Module "Test.A"` is a child module of the `Module "Test"`, the `Module "Test.A.B"` is a child module of the `Module "Test.A"`.

The child module can read all its parent's global variables, the root module can read all global variables from `_G`, beware, the read action will cause the variable be auto-cached, change the global variable in the root module won't affect its children, if you want share some modifable variables, try use a table as container.

The child can't read variables from its brothers or descendants.


### Version Control

No matter how may times call the `Module "Test.A"`, the same module'll be return, so in some scenarios, a moudle could be loaded several times(if embed libs into libs), we need check the version of the module, only the newest module should be loaded to override the old ones:

```lua
require "PLoop"

_ENV = Module "Test" "1.0.12"

_ENV = Module "Test" "1.0.11.12" -- Error: there is an equal or bigger version existed
```

So, the module'll check the version, if it's not newer than the  existed one, an error'll be fired.

If we don't need any error, we should do the check manually:

```lua
local module = Module "Test"
local VER    = "1.011"

if not module:ValidateVersion(VER) then return end

_ENV = module(VER)

function Test() end
```

The version string check works like `"a 1.0.0" < "b 1.0.0" < "v 1.0.0" < "r 1.0.0" < "r 1.1"`, it'll first check the non-num part, then compare each number part.


### Usable features

Like access `_G` in the `_G`, we could use severa features in the module itself:

```lua
_ENV = Module "Test.A.B.C" "v1.0.1"

print(_M)                            -- the module itself
print(_Name)                         -- the module name "C"
print(_FullName)                     -- the module full name "Test.A.B.C"
print(_Parent)                       -- the module parent Module "Test.A.B"
print(_Version)                      -- the module version "v1.0.1"
print(_Parent._Modules["C"] == _M)   -- Use _Modules to access children

-- get all children
for name, mdl in _Parent:GetModules() do
	print(name, mdl)
end
```

The `GetModules` and `ValidateVersion` are methods of the modules, and the `_M`, `_Name`, `_FullName`, `_Parent`, `_Version`, `_Modules` are properties of the modules, we'll learn those in the class chapter.


### Create a new module system

We may create new module systems, like application for web server. To avoid the environment controlls, we can inherit the **System.Module**.

```lua
require "PLoop" (function(_ENV)
	class "Application" { Module }

	Application "Test" (function(_ENV)
	end)

	print(Application("Test") == Module("Test")) -- false
end)
```

So we have a new module system, it won't conflict with the same name module of the **System.Moodule**.
