# namespace

The namespaces are used to organize feature types. Same name features can be saved in different namespaces so there won't be any conflict. Like **System.Form.List** is an UI type, and the **System.Collections.List** is a data container.


## Table of Contents

* [The namespace accessing](#the-namespace-accessing)
* [namespace and environment](#namespace-and-environment)
* [System.Namespace](#systemnamespace)
	* [ExportNamespace](#exportnamespace)
	* [GetNamespace](#getnamespace)
	* [GetNamespaces](#getnamespaces)
	* [GetNamespaceName](#getnamespacename)
	* [GetNamespaceForNext](#getnamespacefornext)
	* [IsAnonymousNamespace](#isanonymousnamespace)
	* [SaveNamespace](#savenamespace)
	* [SaveAnonymousNamespace](#saveanonymousnamespace)
	* [SetNamespaceForNext](#setnamespacefornext)
	* [Validate](#validate)


## The namespace accessing

All namespaces must have one parent namespace except the **PLoop**, its the root namespaces of all others, it's saved to the `_G` when use `require "PLoop"`.

A namespace can only be accessed through its parent like

```lua
require "PLoop"

print(PLoop.System)
print(PLoop.System.Collections)
print(PLoop.System.Collections.List)
```

**PLoop** won't save the namespaces to the `_G` unless use the **import** keyword.

```lua
require "PLoop"

print(System) -- nil

import "System"

print(System) -- System
print(Collections) -- System.Collections

print(List) -- nil

import "System.Collections"

print(XList(100):Sum()) -- 5050
```


## namespace and environment

Unlike the `_G`, it's more simple to use the **PLoop**'s environment system to use namespaces:

```lua
require "PLoop" (function(_ENV)
	print(Collections) -- System.Collections
	print(XList(100):Sum()) -- 5050
end)
```

There are three global namespaces registered to the environment system by the **PLoop** lib, it can be accessed by any private environment:

* System                - So we can use types like Logger, Date, Recycle and etc
* System.Collections    - So we can use types like List, XList, Dictionary, XDictionary and etc
* System.Threading      - So we can use `__Async__`, `__Iterator__` and etc

Besides the global environment, we can use two keywords to manipulate the namespaces for the environment:

* namespace(path)       - set the environment's namespace, so all types defined in the environment will be saved to the namespace
* import(path)          - import a namespace to the environment, so the environment can access the namespace and its sub-namespaces directly.

The environment can import unlimited namespaces, but can only have one namespace as its own.

```lua
require "PLoop"

PLoop(function(_ENV)
	namespace "System.Serialization"

	print(ISerializable) -- System.Serialization.ISerializable

	namespace "Test"

	print(ISerializable) -- nil, the environment don't cache for definition phase

	import "System.Serialization"
	print(ISerializable) -- System.Serialization.ISerializable
end)
```

All types like classes defined in the environment will be saved as child namespace of the environment's namespace. So we can share those types between environments.

```lua
require "PLoop"

PLoop(function(_ENV)
	class "A" {}

	namespace "Test" -- create and set the environment's namespace

	class "B" {} -- B is saved as Test.B
end)

PLoop(function(_ENV)
	import "Test" -- import the Test namespace

	print(A) -- nil
	print(Test) -- Test
	print(B) -- Test.B
end)

```

When an environment try to fetch value from the namespace system, it'll follow those orders:

1. Check the environment's namespace, try to match the namespace's name or its child namespace's name. If nil go to step 2, otherwise return the namespace or its child.

2. Check the environment's imported namespaces, try to match those namespace's name or their child namespace's name. If nil go to step 3, otherwise return the namespace or its child.

3. Check the global namespaces like **System**, try to match the namespace's name or its child namespace's name. If not nil then return the namespace or its child.

If no value can be fetched from the namespaces, the environment'll try its parent environment or the `_G`.


## System.Namespace

The **System.Namespace** is the proxy created from the **namespace** prototype. It contains all features of the **namespace** system.


### ExportNamespace

Export a namespace and its children to an environment

* Format: (env, ns[, override][, stack])
* Params:
	* env           - the environment
	* ns            - the namespace
	* override      - whether override the existed value in the environment
	* stack         - the stack level


### GetNamespace

Get the namespace by path

* Format: ([root, ]path)
* Params:
	* root          - the root namespace
	* path          - string, the namespace path
* Return:
	* ns            - the namespace

```lua
require "PLoop" (function(_ENV)
	-- System.Collections
	print(Namespace.GetNamespace("System.Collections"))

	-- System.Collections.List
	print(Namespace.GetNamespace(System, "Collections.List"))
end)
```


### GetNamespaces

Get the sub-namespaces

* Params:
	* ns            - the namespace
* Return:
	* iter          - function, the iterator
	* ns            - the namespace

```lua
require "PLoop" (function(_ENV)
	for name, ns in Namespace.GetNamespaces(System) do
		print(name, ns)
	end
end)
```


### GetNamespaceName

Get the namespace's path or name

* Format: (ns[, onlyname])
* Params:
	* ns            - the namespace
	* onlyname      - only the name
* Return:
	* name          - string, the path or the name of the namespace


### GetNamespaceForNext

Get the namespace for next generated type, should only be used by the system.

* Return:
	* ns            - the namespace for the next generated type


### IsAnonymousNamespace

Whether the target is anonymous namespace

* Params:
	* ns            - the namespace
* Return:
	* isanony       - boolean, true if the namespace has no name


### SaveNamespace

Save feature to the namespace as the sub-namespace.

* Format: ([root, ]path, feature[, stack])
* Parmas:
	* root          - the root namespace
	* path          - string, the namespace path
	* feature       - the feature, must be table or userdata, normally data types
	* stack         - the stack level


### SaveAnonymousNamespace

Save anonymous namespace, anonymous namespace also can be used as new root of another namespace tree.

* Format: (feature[, stack])
* Parmas:
	* feature       - the feature, must be table or userdata, normally data types
	* stack         - the stack level


### SetNamespaceForNext

Set the namespace for next generated type

* Format: (ns[, stack])
* Parmas:
	* ns            - the namespace for next generated type
	* stack         - the stack level


### Validate

Whether the target is a namespace

* Params:
	* target        - the target to validate
* Return:
	* target        - nil if the target isn't a namespace
