# enum

the enumeration is a validatable data type consisting of a set of named values called elements, The enumerator names are usually identifiers that behave as constants.


## Table of Contents

* [The creation of an enum](#the-creation-of-an-enum)
* [The auto index enum](#the-auto-index-enum)
* [Flags enum](#flags-enum)
* [The re-definition of the enum](#the-re-definition-of-the-enum)
* [The default value of the enum type](#the-default-value-of-the-enum-type)
* [System.Enum](#systemenum)
	* [GetDefault](#getdefault)
	* [GetEnumValues](#getenumvalues)
	* [IsFlagsEnum](#isflagsenum)
	* [IsImmutable](#isimmutable)
	* [IsSealed](#issealed)
	* [IsSubType](#issubtype)
	* [Parse](#parse)
	* [ValidateFlags](#validateflags)
	* [ValidateValue](#validatevalue)
	* [Validate](#validate)
* [Define attribute for the enum](#define-attribute-for-the-enum)


## The creation of an enum

To define an enum within the **PLoop**, the syntax is

```lua
enum "name" { -- key-value pairs }
```

In the table, for each key-value pair:

* if the key is string, the key would be used as the element's name and the value is the element's value.
* If the key is a number and the value is string, the value would be used as both the element's name and value.
* the other key-value pairs will be ignored.

```lua
require "PLoop"(function(_ENV)
	enum "Test" { "Day", Time = "time", [2] = 100, }

	print(Test.Day) -- Day
	print(Test.Time) -- time
	print(Test[2]) -- nil
end)
```


** The usage of the enum

For an enum type, we can use `enumeration[elementname]` to fetch the enum value from name, and use `enumeration(value)` to fetch the element name from value.

We also can use the element name directly where the enum is defined or imported.

Here is an example :

```lua
require "PLoop"

PLoop(function(_ENV)
	namespace "TestNS"

	enum "Direction" { North = 1, East = 2, South = 3, West = 4 }

	print(Direction.South) -- 3
	print(Direction.NoDir) -- nil
	print(Direction(3)) -- South

	print(East) -- 2
end)

PLoop(function(_ENV)
	import "TestNS.Direction"

	print(South) -- 3
end)
```


## The auto index enum

In many conditions, we create enumerations with indexed values:

```lua
require "PLoop"

PLoop(function(_ENV)
	enum "Direction" {
		North = 1,
		East  = 2,
		South = 3,
		West  = 4,
	}
end)
```

If we have many elements of an enum type, it's hard to always keep or change those values.

The `__AutoIndex__` attribute can be used to simple the jobs:

```lua
require "PLoop"

PLoop(function(_ENV)
	__AutoIndex__{ North = 1, South = 5 }
	enum "Direction" {
		"North",
		"East",
		"South",
		"West",
	}

	print(East) -- 2
	print(West) -- 6
end)
```

The `__AutoIndex__` attribute will give each element an auto-increase index based on the config tables.


## Flags enum

Another special enum is the flags enumeration type, the element value should be 2^n(0 is also allowed), so the element value can be used together :

```lua
require "PLoop"

PLoop(function(_ENV)
	__Flags__()
	enum "Days" {
		"SUNDAY",
		"MONDAY",
		"TUESDAY",
		"WEDNESDAY",
		"THURSDAY",
		"FRIDAY",
		"SATURDAY",
	}

	v = SUNDAY + MONDAY + FRIDAY

	-- SUNDAY  1
	-- MONDAY  2
	-- FRIDAY  32
	for name, val in Days(v) do
		print(name, val)
	end

	print(Enum.ValidateFlags(MONDAY, v))   -- true
	print(Enum.ValidateFlags(SATURDAY, v)) -- false
end)
```


## The re-definition of the enum

* If the enum type is not sealed, redefine it will erase the previous definitions.

```lua
require "PLoop" (function(_ENV)
	enum "Test" { "A", "B" }

	print(Test.A) -- A

	enum "Test" { "C" }

	print(Test.A) -- nil
end)
```

* If the enum type is sealed, redefine it will keep the previous definitions.

```lua
require "PLoop" (function(_ENV)
	__Sealed__()
	enum "Test" { "A", "B" }

	print(Test.A) -- A

	enum "Test" { "C" }

	print(Test.A) -- A
	print(Test.C) -- C
end)
```


## The default value of the enum type

As a basic data type, the enum types can have default values :

```lua
require "PLoop" (function(_ENV)
	__Default__("North") -- the element's name or value
	enum "Direction" { North = 1, East = 2, South = 3, West = 4 }

	class "Loc" { Dir = Direction }

	print(Loc().Dir == Direction.North) -- true
end)
```

The default value of the enum type is used by other system as default value, like the property, the struct member field and etc.


## System.Enum

The **System.Enum** is the proxy created from the **enum** prototype. It contains all features of the **enum** system. (The APIs used by System won't be introduced, but you can find them easily in the Prototype.lua.)


### GetDefault

Get the default value from the enumeration.

* Params:
	* enum          - the target enumeration
* Return:
	* default       - the default value


### GetEnumValues

Get the elements from the enumeration

* Params:
	* enum          - the target enumeration
* Return:
	* iter          - function, the iterator
	* enum          - the target enumeration

```lua
require "PLoop" (function(_ENV)
	enum "Test" { "A", "B", C = 3, "D" }

	for name, value in Enum.GetEnumValues(Test) do
		print(name, value)
	end
end)
```


### IsFlagsEnum

Whether the enum type is a flags enum

* Params:
	* enum          - the target enumeration
* Return:
	* isflags       - boolean, true if the enum type is a flags enum


### IsImmutable

Whether the enum's value is immutable through the validation, always true.

* Params:
	* enum          - the target enumeration
* Return:
	* isImmutable   - always true
	* isAlways   	- always true


### IsSealed

Whether the enumeration is sealed, so can't be re-defined

* Params:
	* enum          - the target enumeration
* Return:
	* isSealed      - boolean, true if the enum type is sealed


### IsSubType

Whether the enumeration is sub-type of others, always false.

* Params:
	* enum          - the target enumeration
	* super         - the super type
* Return:
	* isSubType     - boolean, always false


### Parse

Parse the element value to element name

* Params:
	* enum          - the target enumeration
	* value         - the value
* Return Format:
	* (name)
	* (iter, enum)
* Return:
	* name          - if the enum is not flags enum, return the element name for the value
	* iter          - if the enum is flags enum, return the iterator
	* enum          - the taget enumeration


### ValidateFlags

Whether the check value contains the target flag value

* Params:
	* target        - the target value, should only be 2^n
	* check         - the check value
* Return
	* contains      - boolean, true if the check value contains the target value

```lua
require "PLoop" (function(_ENV)
	print(print(Enum.ValidateFlags(4, 7))) -- true, 7 = 1 + 2 + 4
end)
```

### ValidateValue

Whether the value is the enumeration's element's name or value

* Format: (enum, value[, onlyvalid])
* Params:
	* enum          - the target enumeration
	* value         - the enum's name or value
	* onlyvalid     - if true use true instead of the error message
* Return:
	* value         - the element value, nil if not pass the validation
	* errormsg      - the error message if not pass


### Validate

Whether the value is an enumeration

* Params:
	* target        - the target
* Return
	* target        - nil if not pass the validation


For a valid data type, its prototype must provide the **IsImmutable**, **IsSubType** and **ValidateValue** methods. So the enum is a valid data type.


## Define attribute for the enum

Normally attribute like `__AutoIndex__` is used to modify the enum's definition table, here is an implementation  to the `__AutoIndex__`

```lua
require "PLoop" (function(_ENV)
	class "__AutoIdx__" (function(_ENV)
		extend "IInitAttribute"

		property "AttributeTarget" { set = false, default = AttributeTargets.Enum }

		-- target       - the enum
		-- definition   - table, the enum's definition
		function InitDefinition(self, target, targettype, definition)
			local set = self[1]

			local newdef = {}
			local idx = 0

			if set and type(set) ~= "table" then set = nil end

			-- change { "Ann", "Ben" } to { Ann = 1, Ben = 2 }
			for _, name in ipairs(definition) do
				idx = set and set[name] or (idx + 1)
				newdef[name]= idx
			end

			return newdef
		end

		__Arguments__{ RawTable/nil }
		function __new(_, set)
			return { set or false }, true
		end
	end)

	__AutoIdx__{ Ben = 3 }
	enum "Person" { "Ann" , "Ben", "Cook" }

	print(Ann)  -- 1
	print(Cook) -- 4
end)
```