# struct

The structures are validatable data types for basic and complex organized datas and also the data contracts for value validation.

The struct system is focused on the value's validation, to meet all needs, there are four struct types: custom for basic data type, member for complex table data type, array for array data type, dict for hash table types.


## Table of Contents

* [Custom struct](#custom-struct)
	* [Validator](#validator)
	* [Initializer](#initializer)
	* [Inheritance](#inheritance)
	* [System custom structs](#system-custom-structs)
* [Member](#member)
* [Array](#array)
* [Dictionary](#dictionary)
* [Table Style Definition](#table-style-definition)
* [Reduce the validation cost](#reduce-the-validation-cost)
* [Combine type](#combine-type)
* [Sub Type](#sub-type)
* [The re-definition of the struct](#the-re-definition-of-the-struct)
* [Template Struct](#template-struct)
* [Object Validation](#object-validation)
* [System.Struct](#systemstruct)
	* [GetArrayElement](#getarrayelement)
	* [GetBaseStruct](#getbasestruct)
	* [GetComboTypes](#getcombotypes)
	* [GetDefault](#getdefault)
	* [GetDictionaryKey](#getdictionarykey)
	* [GetDictionaryValue](#getdictionaryvalue)
	* [GetErrorMessage](#geterrormessage)
	* [GetMainType](#getmaintype)
	* [GetMember](#getmember)
	* [GetMembers](#getmembers)
	* [GetMethod](#getmethod)
	* [GetMethods](#getmethods)
	* [GetStructCategory](#getstructcategory)
	* [GetTemplate](#gettemplate)
	* [GetTemplateParameters](#gettemplateparameters)
	* [IsImmutable](#isimmutable)
	* [IsObjectAllowed](#isobjectallowed)
	* [IsSubType](#issubtype)
	* [IsSealed](#issealed)
	* [IsStaticMethod](#isstaticmethod)
	* [ValidateValue](#validatevalue)
	* [Validate](#validate)
* [System.Member](#systemmember)
	* [GetType](#gettype)
	* [IsRequire](#isrequire)
	* [GetName](#getname)
	* [GetDefault](#getdefault)


## Custom struct

The custom structs are defined for the basic data types like number, string, boolean and etc.

### Validator

Take the *Number* as an example:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Number" (function(_ENV)
		function Number(value)
			if type(value) ~= "number" then
				return "the %s must be number, got " .. type(value)
			end
		end
	end)

	v = Number(true) -- Error : the value must be number, got boolean
end)
```

The definition body of the struct *Number* is a funciton with `_ENV` as its first parameter, so like using the *PLoop* to process the code, the definition is processed in a private environment created to define the struct type.

Besides the common features come from the **environment** prototype, the private environment still have several special rules for definition, the first rule is for the **validator**:

	It'll check the assignment to the environment, if a function is defined with the struct's name, the function will be used as the struct's **validator**, also you could use `__valid` instead of the struct's name(special for anonymous struct type).

The **validator** is used to validate the input value, if the return value is non-false, that means the target value can't pass the validation, normally the return value should be an error message, the `%s` in the message'll be replaced by name based on where it's used, if the return value is true, the system will generate the error message for it.


### Initializer

In some case, we may need to change the input value to another one, that's done within the **initializer** which is declared like :

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "AnyBool" (function(_ENV)
		function __init(value)
			return value and true or fale
		end
	end)
	print(AnyBool(1)) -- true
end)
```

The function named `__init` is the **initializer**, it's used to modify the target value, if the return value is non-nil, it'll be used as the new value.

We'll see a more usable example for it:

```lua
require "PLoop" (function(_ENV)
	__Arguments__{ Lambda, Any * 0 }
	function Calc(func, ...)
		print(func(...))
	end

	Calc("x,y=>x+y", 1, 11) -- 12
	Calc("x,y=>x*y", 2, 11) -- 22

	Calc("x->x^2", 3) -- Usage: Calc(System.Lambda, [... as System.Any]) - the 1st argument must be a string like 'x,y=>x+y'
end)
```

The **System.Lambda** is used to convert strings like `"x,y=>x+y"` to funciton like `function(x,y) return x+y end`. It's using the **initializer** to do the job.

The custom structs with **initializer** are mutable data types, the custom structs without the **initializer** are immutable data types(you shouldn't change the value(if the value is table) in the validator). So the **Number** is an immutable data type, the **Lambda** is a mutable data type.


### Inheritance

The struct type can have one base struct so it will inherit the base struct's validator and initializer, the base struct's validator and initializer will be called before the struct's own:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Integer" (function(_ENV)
		__base = Number

		local floor = math.floor

		function Integer(value)
			return floor(value) ~= value and "the %s must be integer"
		end
	end)

	v = Integer(true) -- Error : the value must be number, got boolean
	v = Integer(1.23) -- Error : the value must be integer
end)
```

Like the enum, we also can provide a default value to the custom struct since they are normally basic datas:

```lua
require "PLoop"

PLoop(function(_ENV)
	__Default__(0)
	struct "Integer" (function(_ENV)
		__base = Number
		__default = 0 -- also can use this instead of the __Default__

		local floor = math.floor

		function Integer(value)
			return floor(value) ~= value and "the %s must be integer"
		end
	end)

	print(Struct.GetDefault(Integer)) -- 0
end)
```


### System custom structs

There system have provide many fundamental custom struct types like :

Custom Type                   |Description
:-----------------------------|:-----------------------------
**System.Any**                |represents any value
**System.Boolean**            |represents boolean value
**System.String**             |represents string value
**System.Number**             |represents number value
**System.Function**           |represents function value
**System.Table**              |represents table value
**System.Userdata**           |represents userdata value
**System.Thread**             |represents thread value
**System.AnyBool**            |represents anybool value
**System.NEString**           |represents nestring value
**System.RawTable**           |represents rawtable value
**System.Integer**            |represents integer value
**System.NaturalNumber**      |represents natural number value
**System.NegativeInteger**    |represents negative interger value
**System.NamespaceType**      |represents namespace type
**System.EnumType**           |represents enum type
**System.StructType**         |represents struct type
**System.InterfaceType**      |represents interface type
**System.ClassType**          |represents class type
**System.AnyType**            |represents any validation type
**System.Lambda**             |represents lambda value
**System.Callable**           |represents callable value, like function, callable objecct, lambda
**System.Guid**               |represents Guid value


## Member

The member structure represent tables with fixed fields of certain types. Take an example to start:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Location" (function(_ENV)
		x = Number
		y = Number
	end)

	loc = Location{ x = "x" } -- Error: Usage: Location(x, y) - x must be number
	loc = Location(100, 20)
	print(loc.x, loc.y) -- 100  20
end)
```

We already know the definition environment is special designed for the struct types, so if it found an assignment with a string key and the value is a validatable data type, the assignment will be consumed as the creation of a new member, the member type will be used to validate the value fields.

The member sturt can also be used as value constructor(and only the member struct can be used as constructor), the argument order is the same order as the declaration of it members.

The `x = Number` is the simplest way to declare a member to the struct, but there are other details to be filled in, here is the formal version:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Location" (function(_ENV)
		member "x" { type = Number, require = true }
		member "y" { type = Number, default = 0    }
	end)

	loc = Location{} -- Error: Usage: Location(x, y) - x can't be nil
	loc = Location(100)
	print(loc.x, loc.y) -- 100  0
end)
```

The **member** is a keyword can only be used in the definition body of a struct, it need a member name and a table contains several settings for the member(the field is case ignored, and all optional):

Field             |Description
:-----------------|:--------------
type              |the member's type, must be a validatable data type.
require           |boolean, Whether the member can't be nil.
default           |the default value of the member, the type's default value may be used if not provided.

The member struct also support the validator and initializer :

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "MinMax" (function(_ENV)
		member "min" { Type = Number, Require = true }
		member "max" { Type = Number, Require = true }

		function MinMax(val)
			return val.min > val.max and "%s.min can't be greater than %s.max"
		end
	end)

	v = MinMax(100, 20) -- Error: Usage: MinMax(min, max) - min can't be greater than max
end)
```

Since the member struct's value are tables, we also can define struct methods that would be saved to those values:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Location" (function(_ENV)
		member "x" { Type = Number, Require = true }
		member "y" { Type = Number, Default = 0    }

		function GetRange(val)
			return math.sqrt(val.x^2 + val.y^2)
		end
	end)

	print(Location(3, 4):GetRange()) -- 5
end
```

We can also declare static methods that can only be used by the struct itself(also for the custom struct):

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Location" (function(_ENV)
		member "x" { Type = Number, Require = true }
		member "y" { Type = Number, Default = 0    }

		__Static__()
		function GetRange(val)
			return math.sqrt(val.x^2 + val.y^2)
		end
	end)

	print(Location.GetRange{x = 3, y = 4}) -- 5
end)
```

The `System.__Static__` is an attribute, it's used here to declare the next defined method is a static one.

In the previous example, we can give the custom struct a default value, now we'll see how the default value is used:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Number" (function(_ENV)
		__default = 0

		function Number(value)
			return type(value) ~= "number" and "the %s must be number"
		end
	end)

	struct "Location" (function(_ENV)
		x = Number
		y = Number
	end)

	loc = Location()
	print(loc.x, loc.y) -- 0  0
end)
```

So the member would use the type's default value as its default value.

The member struct can also have base struct, it will inherit members, non-static methods, validator and initializer, but it's not recommended.

The member struct is immutable only when:

* it has no non-static method
* it has no initializer
* all its member's type are immutable
* all its member has no default value

The system only provide one member struct type:

Member Type                   |Description
:-----------------------------|:-----------------------------
**System.Variable**           |represents variable, normally used with the `__Arguments__` attribute


## Array

The array structure represent tables that contains a list of same type items. Here is an example to declare an array:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Location" (function(_ENV)
		x = Number
		y = Number
	end)

	struct "Locations" (function(_ENV)
		__array = Location
	end)

	v = Locations{ {x = true} } -- Usage: Locations(...) - the [1].x must be number
end)
```

The array structure also support methods, static methods, base struct, validator and initializer.

The array struct type is immutable only when:

* the array struct type has no initializer
* the array struct's element type is immutable

The system only provide one array struct type:

Array Type                    |Description
:-----------------------------|:-----------------------------
**System.Variables**          |represents variables, normally used with the `__Arguments__`


## Dictionary

The dictionary is used to validate the hash table values. It use `__key` to specific the type of the key, and `__value` to specific the type of the value.

```lua
require "PLoop" (function(_ENV)
	struct "NameID" (function(_ENV)
		__key = String
		__value = Number
	end)

	class "A" { Name = struct { z = NameID } }

	-- xxx.lua:10: the value in Name.z must be number, got string
	A().Name = { z = {  x = "A" } }
end)
```

The dict structure also support methods, static methods, base struct, validator and initializer.

The dict struct type is immutable only when:

* the dict struct type has no initializer
* the dict struct's key type is immutable
* the dict struct's value type is immutable


## Table Style Definition

To simplify the definition of the struct, table can be used instead of the function as the definition body.

```lua
require "PLoop"

PLoop(function(_ENV)
	-- Custom struct
	__Sealed__()
	struct "Number" {
		__default = 0,  -- The default value

		-- the function with number index would be used as validator
		function (val) return type(val) ~= "number" end,

		-- Or you can clearly declare it
		__valid = function (val) return type(val) ~= "number" end,
	}

	struct "AnyBool" {
		__init = function(val) return val and true or false end,
	}

	-- Member struct
	struct "Location" {
		-- Like use the member keyword, just with a name field
		{ name = "x", type = Number, require = true },
		{ name = "y", type = Number, require = true },

		-- Define methods
		GetRange = function(val) return math.sqrt(val.x^2 + val.y^2) end,
	}

	-- Array struct
	-- A valid type with number index, also can use the __array as the key
	struct "Locations" { Location }

	-- Dict struct
	-- The key and value are all types
	struct "NameID" { [String] = Number}
end)
```

## Reduce the validation cost

Let's return the first struct **Number**, the error message is generated during runtime, and in **PLoop** there are many scenarios we only care whether the value match the struct type, so we only need validation, not the error message(the overload system use this technique to choose function).

The validator can receive 2nd parameter which indicated whether the system only care if the value is valid, so we can avoid the generate of new strings when we only need validate it like:

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Number" (function(_ENV)
		function Number(value, onlyvalid)
			if type(value) ~= "number" then return onlyvalid or "the %s must be number, got " .. type(value) end
		end
	end)

	-- The API to validate value with types (type, value, onlyvald)
	print(Struct.ValidateValue(Number, "test", true))  -- nil, true
	print(Struct.ValidateValue(Number, "test", false)) -- nil, the %s must be number, got string
end)
```

Also you can just return true so the system'll take care of the rest part.


## Combine type

If your value could be two or more types, you can combine those types like :

```lua
require "PLoop"

PLoop(function(_ENV)
	-- nil, the %s must be value of System.Number | System.String
	print(Struct.ValidateValue(Number + String, {}, false))
end)
```

You can combine any validatable types like enums, structs, interfaces or classes, it'd generate a struct type, the same combination will only return the generated struct type.


## Sub Type

If you need the value to be a struct who is a sub type of another struct, (a struct type is a sub type of itself), you can create it like `- Number` :

```lua
require "PLoop"

PLoop(function(_ENV)
	struct "Integer" { __base = Number, function(val) return math.floor(val) ~= val end }
	print(Struct.ValidateValue( - Number, Integer, false)) -- Integer
end)
```

You can use the `-` operation on any validatable types with **IsSubType** API, include the struct, interface and class.


## The re-definition of the struct

* If the struct type is not sealed, redefine it will erase the previous definitions.

```lua
require "PLoop" (function(_ENV)
	struct "Test" { function() end }

	print(Struct.GetStructCategory(Test) == StructCategory.CUSTOM)

	struct "Test" { Number }

	print(Struct.GetStructCategory(Test) == StructCategory.ARRAY)
end)
```

* If the struct type is sealed, we can't re-define it.

```lua
require "PLoop" (function(_ENV)
	__Sealed__()
	struct "Test" { function() end }

	struct "Test" { Number } -- Usage: struct.BeginDefinition(structure[, stack]) - The Test is sealed, can't be re-defined
end)
```


## Template Struct

Since we use the function as the definition body, we also can provide more informations to the definition function.

That would make the struct type as a template struct type. To restrict the template parameters, the **PLoop** use `__Arguments__` attribute to declare the struct type as template.

```lua
require "PLoop"

PLoop(function(_ENV)
	__Arguments__{ Number, Number }
	struct "FixString" (function(_ENV, min, max)
		__base = String

		function __valid(val)
			return (#val > max or #val < min) and "the %s length must between [" .. min .. "," .. max .. "]"
		end
	end)

	V1_3 = FixString[{1, 3}]

	-- true, we get the same type with the same arguments
	print(V1_3 == FixString[{1, 3}])

	-- we can get the prototype and the template parameters
	-- FixString
	print(Struct.GetTemplate(V1_3))

	-- 1    3
	print(Struct.GetTemplateParameters(V1_3))

	-- Error:  the value length must between [1,3]
	a = V1_3("Hello")
end)
```

The template types *FixString* can't be used directly since we don't give the default template parameters, we could do that by add those parameters after the declaration of the `__Arguments__` attribute:

```lua
require "PLoop"

PLoop(function(_ENV)
	__Arguments__{ Number, Number }(1, 4)
	struct "FixString" (function(_ENV, min, max)
		__base = String

		function __valid(val)
			return (#val > max or #val < min) and "the %s length must between [" .. min .. "," .. max .. "]"
		end
	end)

	-- Error: the value length must between [1,4]
	a = FixString("Hello")
end)
```


## Object Validation

Normally, the table has meta-method setting can't pass the **member struct type**'s validation, it's a simple design and keep the system away from some confusable conditions. But for some case, we only need the value match the structure schema, and don't really care what its, we can use `System.__ObjectAllowed__` attribute to mark the struct type, so it could let the class objects pass its validation:

```lua
require "PLoop" (function(_ENV)
	struct "ColorFloat" {
	    __base = Number,
	    function(val, onlyvalid) if (val < 0 or val > 1) then return onlyvalid or "the %s must between [0, 1]" end end
	}

	__ObjectAllowed__()
	struct "ColorType" {
	    { name = "r",   type = ColorFloat, require = true },
	    { name = "g",   type = ColorFloat, require = true },
	    { name = "b",   type = ColorFloat, require = true },
	    { name = "a",   type = ColorFloat, default = 1 },
	}

	class "Color" (function(_ENV)

	    __Arguments__{ ColorType }
	    function __new(_, color)
	        return color, true
	    end

	    __Arguments__{
	        Variable("r", ColorFloat),
	        Variable("g", ColorFloat),
	        Variable("b", ColorFloat),
	        Variable("a", ColorFloat, true, 1),
	    }
	    function __new(_, r, g, b, a)
	        return { r = r, g = g, b = b, a = a }, true
	    end
	end)

	__Arguments__{ ColorType }
	function Test(color)
		print(color.r, color.g, color.b)
	end

	Test(Color(1, 0, 0))
end)
```

**To allow the objects, the first member of the struct type should be none-table values, otherwise it may cause problems**


## System.Struct

The **System.Struct** is the proxy created from the **struct** prototype. It contains all features of the **struct** system. (The APIs used by System won't be introduced, but you can find them easily in the Prototype.lua.)

### GetArrayElement

Get the array structure's element type

* Params:
	* struct        - the target structure
* Return:
	* type          - the array element's type


### GetBaseStruct

Get the structure's base struct type

* Params:
	* struct        - the target structure
* Return:
	* type          - the base struct


### GetComboTypes

Get the types of a combo struct type

* Params:
	* struct        - the target structure
* Returns:
	* type1         - the first type of the combo
	* type2         - the second type of the combo

```lua
require "PLoop" (function(_ENV)
	-- System.Number | System.String	System.Boolean
	print(Struct.GetComboTypes(Number + String + Boolean))
end)
```


### GetDefault

Get the default value from the structure type.

* Params:
	* struct        - the target structure
* Return:
	* default       - the default value


### GetErrorMessage

Generate an error message with template and target, normally used by the system to generate the final error message

* Params:
	* template      - the error message template, normally generated by type validation
	* target        - the target string, like "value"
* Return:
	* message       - the error message

```lua
require "PLoop" (function(_ENV)
	-- the Person.name must be number
	print(Struct.GetErrorMessage("the %s must be number", "Person.name"))
end)
```

### GetMainType

Get the master type of the generated struct for its sub-types.

* Params:
	* target        - the target structure
* Return:
	* type          - the master type

```lua
require "PLoop" (function(_ENV)
	-- System.Number
	print(Struct.GetMainType(-Number))
end)
```


### GetMember

Get the member of the structure with given name, the result is an object, see **System.Member** for more details.

* Params:
	* struct        - the target structure
	* name          - the member's name
* Return:
	* member        - the member object


### GetMembers

Get the members of the structure, see **System.Member** for more details.

* Params:
	* struct        - the target structure
* Return:
	* iter          - function, the iterator
	* struct        - the target structure


### GetMethod

Get the method of the structure with given name

* Params:
	* struct        - the target structure
	* name          - the method's name
* Return:
	* method        - the method
	* isstatic      - boolean, whether the method is static


### GetMethods

Get all the methods of the structure

* Params:
	* struct        - the target structure
* Return:
	* iter          - function, the iterator
	* struct        - the target structure

```lua
require "PLoop" (function(_ENV)
	for name, func, isstatic in Struct.GetMethods(System.Variable) do
		print(name, isstatic)
	end
end)
```


### GetStructCategory

Get the struct category of the structure

* Params:
	* struct        - the target structure
* Return:
	* category      - string, the structure's category: CUSTOM|ARRAY|MEMBER


### GetTemplate

Get template struct type of the target struct

* Params:
	* struct        - the target structure
* Return:
	* template      - the template struct, maybe itself


### GetTemplateParameters

Get the template parameters

* Params:
	* struct        - the target structure
* Return:
	* ...           - the paramter list


### IsImmutable

Whether the struct type is immutable

* Params:
	* struct        - the target structure
* Return:
	* immutable 	- true if the type is immutable
	* isAlways   	- true if the type is always immutable, only when the type is immutable and sealed


## IsObjectAllowed

Whether the strut type allow objects pass its validation

* Params:
	* struct        - the target structure
* Return:
	* allowed       - true if objects can pass the type's validation


### IsSubType

Whether a structure use the other as its base structure

* Params:
	* struct        - the target structure
	* base          - the base structure
* Return:
	* issubtype - true if the structure use the base structure as base


### IsSealed

Whether the structure is sealed, can't be re-defined

* Params:
	* struct        - the target structure
* Return:
	* sealed        - true if the structure is sealed


### IsStaticMethod

Whether the structure's given name method is static

* Params:
	* struct        - the target structure
	* name          - the method's name
* Return:
	* isstatic      - boolean, whether the method is static


### ValidateValue

Validate the value with a structure

* Format: (structure, value[, onlyvalid])
* Params:
	* structure     - the target structure
	* value         - the value
	* onlyvalid     - true if only validate the value, no value modifiy(The initializer and object methods won't be applied)
* Return:
	* value         - the validated value
	* errormsg      - the error message if not pass


### Validate

Whether the value is a struct type

* Params:
	* target        - the target
* Return
	* target        - nil if not pass the validation


## System.Member

We will get member object from the **Struct.GetMemeber** and **Struct.GetMembers** API, we also have a **System.Member** to get those member's informations.

Here is an example :

```lua
require "PLoop" (function(_ENV)
	for i, member in Struct.GetMembers(Variable) do
		print(Member.GetName(member), Member.GetType(member), Member.IsRequire(member), Member.GetDefault(member))
	end
end)
```

### GetType

Get the type of the member

* Params:
	* member        - the member
* Return:
	* type          - the member's type


### IsRequire

Whether the member's value is required

* Params:
	* member        - the member
* Return:
	* required      - true if the member's value is required


### GetName

Get the name of the member

* Params:
	* member        - the member
* Return:
	* name          - the member's name


### GetDefault

Get the default value of the member

* Params:
	* member        - the member
* Return:
	* default       - the member's default value

