# attribute

The **attribute** is a prototype used to provide the attribute system.

The attributes are used to bind informations to features, or used to modify those features directly.

The attribute system is designed to given description labels for codes, so the **PLoop** or other libs can use those labels to modify or extend the codes.

We can register attributes before we define features, and use those attributes to modify the features definition, apply changes to the feature or attach data to them.

**You can skip this chapter if you don't want create your own attribute types.**
**You can skip to the [Define attribute types](#define-attribute-types) if you only want to know how the create attribute types**

The attribute can be any userdata or table with attribute usage informations, normally they are objects created from classes.


## Table of Contents

* [Attribute Usages](#attribute-usages)
* [System.AttributeTargets](#systemattributetargets)
* [System.Attribute](#systemattribute)
	* [RegisterTargetType](#registertargettype)
	* [Register](#register)
	* [HaveRegisteredAttributes](#haveregisteredattributes)
	* [SaveAttributes](#saveattributes)
	* [InheritAttributes](#inheritattributes)
	* [InitDefinition](#initdefinition)
	* [ToggleTarget](#toggletarget)
	* [ApplyAttributes](#applyattributes)
	* [AttachAttributes](#attachattributes)
	* [GetAttachedData](#getattacheddata)
	* [HasAttachedData](#hasattacheddata)
	* [GetAttributeTargets](#getattributetargets)
	* [GetAttributeTargetOwners](#getattributetargetowners)
	* [IndependentCall](#independentcall)
	* [ReleaseTargetAttributes](#releasetargetattributes)
	* [Unregister](#unregister)
* [Define attribute types](#define-attribute-types)
	* [System.IAttribute](#systemiattribute)
	* [System.IInitAttribute](#systemiinitattribute)
	* [System.IApplyAttribute](#systemiapplyattribute)
	* [System.IAttachAttribute](#systemiattachattribute)
* [System Attributes](#system-attributes)
	* [`__Abstract__`](#__abstract__)
	* [`__AutoCache__`](#__autocache__)
	* [`__AnonymousClass__`](#__anonymousclass__)
	* [`__AutoIndex__`](#__autoindex__)
	* [`__Arguments__`](#__arguments__)
	* [`__Base__`](#__base__)
	* [`__Default__`](#__default__)
	* [`__Delegate__`](#__delegate__)
	* [`__EventChangeHandler__`](#__eventchangehandler__)
	* [`__Final__`](#__final__)
	* [`__Flags__`](#__flags__)
	* [`__Get__`](#__get__)
	* [`__Indexer__`](#__indexer__)
	* [`__Namespace__`](#__namespace__)
	* [`__NoNilValue__`](#__nonilvalue__)
	* [`__NoRawSet__`](#__norawset__)
	* [`__ObjectAttr__`](#__objectattr__)
	* [`__ObjFuncAttr__`](#__objfuncattr__)
	* [`__ObjectSource__`](#__objectsource__)
	* [`__Require__`](#__require__)
	* [`__Return__`](#__return__)
	* [`__Sealed__`](#__sealed__)
	* [`__Set__`](#__set__)
	* [`__SingleVer__`](#__singlever__)
	* [`__Static__`](#__static__)
	* [`__Super__`](#__super__)
	* [`__SuperObject__`](#__superobject__)
	* [`__Throwable__`](#__throwable__)


## Attribute Usages

The attribute usages are fixed name fields, methods or properties of the attribute:

* InitDefinition        - A method used to modify the target's definition or init the target before it load its definition, and its return value will be used as the new definition for the target if existed.
	* Parameters :
		* attribute     - the attribute
		* target        - the target like class, method and etc
		* targettype    - the target type, that's a flag value registered by types.
		* definition    - the definition of the target.
		* owner         - the target's owner, it the target is a method, the owner may be a class or interface that contains it.
		* name          - the target's name, like method name.
	* Returns :
		* definiton     - the return value will be used as the target's new definition.

* ApplyAttribute        - A method used to apply the attribute to the target. the method would be called after the definition of the target. The target still can be modified.
	* Parameters :
		* attribute     - the attribute
		* target        - the target like class, method and etc
		* targettype    - the target type, that's a flag value registered by the target type.
		* manager       - the definition manager of the target, normally the definition environment of the target, it's a little dangerous to use the definition environment directly, but very useful.
		* owner         - the target's owner, it the target is a method, the owner may be a class or interface that contains it.
		* name          - the target's name, like method name.

* AttachAttribute       - A method used to generate attachment to the target such as runtime document, map to database and etc. The method would be called after the definition of the target. The target can't be modified.
	* Parameters :
		* attribute     - the attribute
		* target        - the target like class, method and etc
		* targettype    - the target type, that's a flag value registered by the target type.
		* owner         - the target's owner, it the target is a method, the owner may be a class or interface that contains it.
		* name          - the target's name, like method name.
	* Returns :
		* attach        - the return value will be used as attachment of the attribute's type for the target.

* AttributeTarget       - Default 0 (all types). The flags that represents the type of the target like class, method and other features.

* Inheritable           - Default false. Whether the attribute is inheritable.

* Overridable           - Default true. Whether the attribute's saved data is overridable.

* Priority              - Default 0. The attribute's priority, the bigger the first to be applied.

* SubLevel              - Default 0. The priority's sublevel, for attributes with same priority, the bigger sublevel the first be applied.


To fetch the attribute usages from an attribute, take the *ApplyAttribute* as an example, the system will first use `attr["ApplyAttribute"]` to fetch the value, since the system don't care how it's provided, field, property, `__index` all should works.

If the attribute don't provide attribute usage, the default value will be used.

Although the attribute system is designed without the type requirement, it's better to define them by creating classes extend **System.IAttribute**.


## System.AttributeTargets

The target's type are used through all attribute usages, we can register any target types as we need, but the system has already provided enough target types:

* System.AttributeTargets.All       - the attribute type represents all
* System.AttributeTargets.Function  - the attribute type represents function
* System.AttributeTargets.Namespace - the attribute type represents namespace
* System.AttributeTargets.Enum      - the attribute type represents enum
* System.AttributeTargets.Struct    - the attribute type represents struct
* System.AttributeTargets.Member    - the attribute type represents member
* System.AttributeTargets.Method    - the attribute type represents method
* System.AttributeTargets.Interface - the attribute type represents interface
* System.AttributeTargets.Class     - the attribute type represents class
* System.AttributeTargets.Object    - the attribute type represents object
* System.AttributeTargets.Event     - the attribute type represents event
* System.AttributeTargets.Property  - the attribute type represents property


## System.Attribute

The **attribute** prototype isn't provided directly, normally we use the **System.Attribute** to access all features of the the attribute system. We'll know the details of the attribute system from those APIS.

As an example, we'll create a module prototype and we can apply attributes on its functions.

```lua
require "PLoop"

PLoop(function(_ENV)
	module = prototype { ... } -- we'll fill the settings later

	obj = prototype.NewObject(module)

	-- We can use the attribute to run the function as an iterator
	__Iterator__()
	function obj:GetIter()
		for i = 1, 10 do
			coroutine.yield(i)
		end
	end
end)
```


### RegisterTargetType

Register an attribute target type, like function, class, interface and etc.

The attribute system has no knowledge about the target features, so before any system use the attribute system, it must register the target type to the attribute system.

* Format: (name[, stack])
* Params:
	* name          - the target type name, like 'class', 'function'
	* stack         - the stack level used for error message
* Return
	* flag          - number, the target type flag number

```lua
require "PLoop"

PLoop(function(_ENV)
	-- 2048
	print(Attribute.RegisterTargetType("MyFeature"))
end)
```

There are many system defined target types in the **PLoop**, so the flag number for the new target type is 2048 now.

The `__Iterator__` attribute only works for two target type :
	* **System.AttributeTargets.Method**
	* **System.AttributeTargets.Function**

So we don't need to create our own attribute target type, we'll use the **System.AttributeTargets.Function** for our example.


### Register

Register the attribute to be used by the next feature. Normally, non-system attributes should extend the **System.IAttribute**, those attributes will be registered when they are created. The `__Iterator__` is also extend the **System.IAttribute**, so there is no need to use it directly.

* Format: (attr[, unique][, stack])
* Params:
	* attr          - the attribute to be registered
	* unique        - boolean, true if no same type attribute can be registered
	* stack         - the stack level used for error message


### HaveRegisteredAttributes

Whether there are registered attributes

* Return:
	* hasattr       - boolean, true if have registered attributes

### SaveAttributes

Save the current registered attributes to the target.

* Format: (target, targettype[, stack])
* Params:
	* target        - the attribute target
	* targettype    - the attribute target type
	* stack         - the stack level used for error message

Now, we can start the definition of the module prototype:

```lua
require "PLoop"

PLoop(function(_ENV)
	module = prototype {
		__newindex = function(self, key, value)
			-- Check if the value is function and we have registered attributes
			if type(value) == "function" and Attribute.HaveRegisteredAttributes() then
				-- Save the attributes to the target
				Attribute.SaveAttributes(value, AttributeTargets.Function)
				-- ...
			end
		end,
	}
end)
```


### InheritAttributes

Save the other target's inheritable attributes to the target(like the super class), must be called after the SaveAttributes and before the AttachAttributes.

* Format: (target, targettype, ...)
* Params:
	* target        - the attribute target
	* targettype    - the attribute target type
	* ...           - the other features that may have inheritable attributes


### InitDefinition

Use the registered attributes to init the target's definition

* Format: (target, targettype, definition, [owner], [name][, stack])
* Params:
	* target        - the attribute target
	* targettype    - the attribute target type
	* definition    - the definition of the target
	* owner         - the target's owner
	* name          - the target's name if it has owner
	* stack         - the stack level used for error message
* Return
	* definition    - the new definition if not nil

For our example:

```lua
require "PLoop"

PLoop(function(_ENV)
	module = prototype {
		__newindex = function(self, key, value)
			if type(value) == "function" and Attribute.HaveRegisteredAttributes() then
				Attribute.SaveAttributes(value, AttributeTargets.Function)

				-- Init the definition of the target, the definition is the function itself
				local newdef = Attribute.InitDefinition(value, AttributeTargets.Function, value, self, key)

				-- ...
			end
		end,
	}
end)
```


### ToggleTarget

Toggle the target, save the old target's attributes to the new one

* Params:
	* old           - the old target
	* new           - the new target

For our example, if the new definition is return, we need use the new definition as the value:

```lua
require "PLoop"

PLoop(function(_ENV)
	module = prototype {
		__newindex = function(self, key, value)
			if type(value) == "function" and Attribute.HaveRegisteredAttributes() then
				Attribute.SaveAttributes(value, AttributeTargets.Function)

				-- Init the definition of the target, the definition is the function itself
				local newdef = Attribute.InitDefinition(value, AttributeTargets.Function, value, self, key)

				if newdef ~= value then
					Attribute.ToggleTarget(value, newdef)
					value = newdef
				end

				-- Apply the definition, for function just save it
				rawset(self, key, value)

				-- ...
			end
		end,
	}
end)
```


### ApplyAttributes

Apply the registered attributes to the target before the definition finished. Normally, we still can do some modification for the target or owner within this method.

* Format: (target, targettype, manager, [owner], [name][, stack])
* Params:
	* target        - the attribute target
	* targettype    - the attribute target type
	* manager       - the definition manager of the target
	* owner         - the target's owner
	* name          - the target's name if it has owner
	* stack         - the stack level used for error message


### AttachAttributes

Attach the registered attributes data to the target after the definition finished. Normally, we can't do modification for the target or the owner within this method.

* Format: (target, targettype, [owner], [name][, stack])
* Params:
	* target        - the attribute target
	* targettype    - the attribute target type
	* owner         - the target's owner
	* name          - the target's name if it has owner
	* stack         - the stack level used for error message
* Return
	* attach        - the attachment is not nil

With our example :

```lua
require "PLoop"

PLoop(function(_ENV)
	module = prototype {
		__newindex = function(self, key, value)
			if type(value) == "function" and Attribute.HaveRegisteredAttributes() then
				Attribute.SaveAttributes(value, AttributeTargets.Function)

				-- Init the definition of the target, the definition is the function itself
				local newdef = Attribute.InitDefinition(value, AttributeTargets.Function, value, self, key)

				if newdef ~= value then
					Attribute.ToggleTarget(value, newdef)
					value = newdef
				end

				-- Apply the definition, for function just save it
				rawset(self, key, value)

				-- keep the manager nil, normally it only used by the class, interface and etc
				Attribute.ApplyAttributes(value, AttributeTargets.Function, nil, self, key)

				-- finish the call of the attribute system
				Attribute.AttachAttributes(value, AttributeTargets.Function, self, key)
			end
		end,
	}

	obj = prototype.NewObject(module)

	-- We can use the attribute to run the function as an iterator
	__Iterator__()
	function obj:GetIter()
		for i = 1, 10 do
			coroutine.yield(i)
		end
	end

	-- All works fine now
	for i in obj:GetIter() do
		print(i)
	end
end)
```


### GetAttachedData

Get the attached attribute data of the target

* Format: (attrtype, target[, owner])
* Params:
	* attrtype      - the attribute type
	* target        - the attribute target
	* owner         - the target's owner
* Return
	* attach        - the attached data


### HasAttachedData

Whether the target of attached data of the attribute type

* Format: (attrtype, target[, owner])
* Params:
	* attrtype      - the attribute type
	* target        - the attribute target
	* owner         - the target's owner
* Return
	* hasattach     - true if has the attached data


### GetAttributeTargets

Get all targets have attached data of the attribute, used as iterator

* Params:
	* attrtype      - the attribute type
* Return
	* iter          - the iterator
	* attrtype      - the attribute type


### GetAttributeTargetOwners

Get all target's owners that have attached data of the attribute, used as iterator

* Params:
	* attrtype      - the attribute type
* Return
	* iter          - the iterator
	* attrtype      - the attribute type


### IndependentCall

Call a definition function within a standalone attribute system so it won't use the registered attributes that belong to others. Normally used in attribute's ApplyAttribute or AttachAttribute that need create new features with attributes.

* Format: (definition[, stack])
* Params:
	* definition    - the function to be processed
	* stack         - the stack level used for error message


### ReleaseTargetAttributes

Release the registered attribute of the target, so those attributes can be collected, the attributes are released automatically if the AttachAttribtues is used, so normally we don't need use it.

* Params:
	* target        - the attribute target


### Unregister

Un-register an attribute, must be used before the attribute is saved to target.

* Params:
	* attr          - the attribute to be un-registered


## Define attribute types

It's recommend to define attribute types as classes, you can learn the class system before read this part.


### System.IAttribute

To define an attribute class, we should extend the **System.IAttribute** interface or its extend interfaces :

* **System.IInitAttribute**     represents the interface to modify the target's definitions

* **System.IApplyAttribute**    represents the interface to apply changes on the target(like mark a enum as flags)

* **System.IAttachAttribute**   represents the interface to attach data on the target(binding database table on class)


The **IAttribute** has defined several properties as the attribute usages:

* AttributeTarget   - System.AttributeTargets, default **System.AttributeTargets.All**

* Inheritable       - boolean, whether the attribute is inheritable, default false

* Overridable       - boolean, Whether the attribute's attach data is overridable, default true

* Priority          - System.AttributePriority, the attribute's priority, the higher the first be applied
	* System.AttributePriority.Highest
	* System.AttributePriority.Higher
	* System.AttributePriority.Normal (Default)
	* System.AttributePriority.Lower
	* System.AttributePriority.Lowest

* SubLevel          - number, the attribute priority's sublevel, if two attribute have the same priority, the bigger sublevel will be first applied, default 0


There are three type attributes, the init attributes are called before the definition of the target, the apply attributes are called during the definition of the target and the attach attributes are called after the definition of the target:


### System.IInitAttribute

Those attributes are used to modify the target's definitions, normally used on functions or enums:

```lua
require "PLoop"

PLoop(function(_ENV)
	class "__SafeCall__" (function(_ENV)
		extend "IInitAttribute"

		local function checkret(ok, ...)
			if ok then return ... end
		end

		--- modify the target's definition
		-- @param target        the target
		-- @param targettype    the target type
		-- @param definition    the target's definition
		-- @param owner         the target's owner
		-- @param name          the target's name in the owner
		-- @param stack         the stack level
		-- @return definition   the new definition
		function InitDefinition(self, target, targettype, definition, owner, name, stack)
			return function(...)
				return checkret(pcall(definition, ...))
			end
		end

		property "AttributeTarget" { default = AttributeTargets.Function + AttributeTargets.Method }
	end)

	__SafeCall__()
	function test1()
		return 1, 2, 3
	end

	__SafeCall__()
	function test2(i, j)
		return i/j
	end

	print(test1()) -- 1, 2, 3
	print(test2()) -- nothing
end)
```

the attribute class should extend the **System.IInitAttribute** and define the **InitDefinition** method to modify the target's definitions, for a function, the definition is the function itself, if the method return a new definition, the new will be used. And for the enum, the definition is the table that contains the elements. The init attributes are called before the define process of the target.


### System.IApplyAttribute

Those attributes are used to apply changes on the target, this is normally used by the system attributes for types like class, struct and etc.

Take the `__Sealed__` attribute as an example:

```lua
class "__Sealed__" (function(_ENV)
	extend "IApplyAttribute"

	--- apply changes on the target
	-- @param target        the target
	-- @param targettype    the target type
	-- @param manager       the definition manager of the target
	-- @param owner         the target's owner
	-- @param name          the target's name in the owner
	-- @param stack         the stack level
	function ApplyAttribute(self, target, targettype, manager, owner, name, stack)
		if targettype == AttributeTargets.Enum then
			Enum.SetSealed(target)
		elseif targettype == AttributeTargets.Struct then
			Struct.SetSealed(target)
		elseif targettype == AttributeTargets.Interface then
			Interface.SetSealed(target)
		elseif targettype == AttributeTargets.Class then
			Class.SetSealed(target)
		end
	end

	property "AttributeTarget" { default = AttributeTargets.Enum + AttributeTargets.Struct + AttributeTargets.Interface + AttributeTargets.Class }
end)
```

the attribute should extend the **System.IApplyAttribute** and define the **ApplyAttribute** method. The apply attributes are applied during the define process of the target.

There is a special *manager* parameter, the definition environment of the struct, interface and class'd be passed in as the value, it's a dangerous but useful feature if you want append common type features into the target:

```lua
require "PLoop"

PLoop(function(_ENV)
	class "__Name__" (function(_ENV)
		extend "IApplyAttribute"

		--- apply changes on the target
		-- @param target        the target
		-- @param targettype    the target type
		-- @param manager       the definition manager of the target
		-- @param owner         the target's owner
		-- @param name          the target's name in the owner
		-- @param stack         the stack level
		function ApplyAttribute(self, target, targettype, manager, owner, name, stack)
			if manager then
				Environment.Apply(manager, function(_ENV)
					property "Name" { type = String }
				end)
			end
		end

		property "AttributeTarget" { default = AttributeTargets.Interface + AttributeTargets.Class }
	end)

	__Name__()
	class "A" {}

	A().Name = 123 -- Error: the Name must be string, got number
end)
```

It's a dangerous feature, so only use it if needed.


### System.IAttachAttribute

Those attributes are used to attach attribute datas on the target, also can be used to register the final result to other systems.

```lua
PLoop(function(_ENV)
	class "__DataTable__" (function(_ENV)
		extend "IAttachAttribute"

		--- apply changes on the target
		-- @param target        the target
		-- @param targettype    the target type
		-- @param owner         the target's owner
		-- @param name          the target's name in the owner
		-- @param stack         the stack level
		function AttachAttribute(self, target, targettype, owner, name, stack)
			return self.DataTable
		end

		property "AttributeTarget" { default = AttributeTargets.Class }

		property "DataTable" { type = String }
	end)

	__DataTable__{ DataTable = "Persons" }
	class "Person" {}

	-- Persons
	print(Attribute.GetAttachedData(__DataTable__, Person))
end)
```

the attribute should extend the **System.IAttachAttribute** and defined the **AttachAttribute** method, the return value of the method will be saved, so we can check it later.



## System Attributes

There are many attributes defined in the **System** namespace to modify the type features:

### `__Abstract__`

Used to mark a class as abstract, so it can't be used to generate objects, or used to mark the method, event or property as abstract, so they need(not must) be implemented by child types.

Attribute Targets:
* System.AttributeTargets.Class
* System.AttributeTargets.Method
* System.AttributeTargets.Event
* System.AttributeTargets.Property


### `__AutoCache__`

Used to mark a class, so its object will cache the object method it accessed into itself, so it don't need to fetch it again for next accessing. If the class enabled the super object style and it has overridden methods or features, the auto-cache will be disabled. Be careful, if the method is cached, it can't be updated with the re-definition of the class.

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function()
	__AutoCache__()
	class "A" { Test = function() end}

	o = A()
	print("GET", rawget(o, "Test")) -- GET	nil
	o:Test()
	print("GET", rawget(o, "Test")) -- GET	function: 0x015be630
end)
```


### `__AnonymousClass__`

Used to mark an interface so it'll auto create an anonymous class that extend the interface. So we can use the interface like a class to generate objects. Only init-table can be used as the parameter, or a function if the interface has only one abstract method.

Attribute Targets:
* System.AttributeTargets.Interface

Usage:

```lua
require "PLoop"

PLoop(function()
	__AnonymousClass__()
	interface "ITask" (function(_ENV)
		__Abstract__() function DoTask(self) end

		function Process(self) self:DoTask() end
	end)

	o = ITask(function() print("Hello") end)
	o:Process()     -- Hello
end)
```


### `__AutoIndex__`

Parse the enum definition as auto-index list.

Attribute Targets:
* System.AttributeTargets.Enum

Usage:

```lua
require "PLoop"

PLoop(function()
	__AutoIndex__ { A = 0, C = 10 }
	enum "Test" { "A", "B", "C", "D" }

	print(Test.A, Test.B, Test.C, Test.D) -- 0, 1, 10, 11
end)
```


### `__Arguments__`

Used to provide type validation to functions, type validation and overload for method. Also used to create template struct, interface or class types.

Attribute Targets:
* System.AttributeTargets.Struct
* System.AttributeTargets.Interface
* System.AttributeTargets.Class
* System.AttributeTargets.Function
* System.AttributeTargets.Method

Usage:

```lua
require "PLoop" (function(_ENV)
	-- type validation
	__Arguments__{ Number }
	function test(age)
		print("Age:", age)
	end

	-- Overload
	class "Person" (function(_ENV)
		__Arguments__{ String }
		function SetInfo(self, name)
			self.name = name
		end

		__Arguments__{ Number }
		function SetInfo(self, age)
			self.age = age
		end
	end)

	-- Template
	__Arguments__{ AnyType }( Any )
	class "Array" (function(_ENV, cls)
		__Arguments__{ cls * 0 }
		function __new(_, ...)
			return { ... }, true
		end
	end)
end)

```


### `__Base__`

Set the target struct's base struct.

Attribute Targets:
* System.AttributeTargets.Struct

```lua
require "PLoop" (function(_ENV)
	struct "Number" { function (val) return type(val) ~= "number" and "the %s must be number" end }

	__Base__(Number)
	struct "Integer" { function(val) return math.floor(val) ~= val and "the %s must be integer" end}

	v = Integer(true)    -- Error: the value must be number
	v = Integer(1.3)     -- Error: the value must be integer
end)
```


### `__Default__`

Set a default value to the enum or custom struct

Attribute Targets:
* System.AttributeTargets.Enum
* System.AttributeTargets.Struct
* System.AttributeTargets.Member


### `__Delegate__`

Decorate the target functions.

Attribute Targets:
* System.AttributeTargets.Function
* System.AttributeTargets.Member

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	function decorate(func, ...)
		print("Call", func, ...)
		return func(...)
	end

	__Delegate__(decorate)
	function test() end

	-- Call function: 02E7B1C8  1   2   3
	test(1, 2, 3)
end)
```


### `__EventChangeHandler__`

Bind a function for an event to handle all changes from its generated delegates.

Attribute Targets:
* System.AttributeTargets.Event


### `__Final__`

Set a class or interface as final, so they can't be inherited or extended by other types. Also can be used to mark the method, event and property as final, so they have the biggest priority to be inherited.

Attribute Targets:
* System.AttributeTargets.Class
* System.AttributeTargets.Interface
* System.AttributeTargets.Method
* System.AttributeTargets.Event
* System.AttributeTargets.Property


### `__Flags__`

Mark the enumeration as flags enumeration.

Attribute Targets:
* System.AttributeTargets.Enum


### `__Get__`

Set the get modifier to a property.

Attribute Targets:
* System.AttributeTargets.Property


### `__Indexer__`

Mark the property as an indexer property.

Attribute Targets:
* System.AttributeTargets.Property


### `__Namespace__`

Set the namespace for the next created type

Attribute Targets:
* System.AttributeTargets.All

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	namespace "Test"

	__Namespace__ "MyNS"
	class "A" {}

	print(A)   -- MyNS.A
end)
```


### `__NoNilValue__`

Set the class's objects so access non-existent fields on them will be denied.

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	__NoNilValue__()
	class "A" {}

	o = A()
	v = o.age -- Error: The object don't have any field that named "age"
end)
```

Also you can mark a class so its object's fields can be accessed no matter how the platform settings is applied

```lua
PLOOP_PLATFORM_SETTINGS = { OBJECT_NO_RAWSEST = true, OBJECT_NO_NIL_ACCESS = true }

require "PLoop"

PLoop(function(_ENV)
	__NoNilValue__(false)
	class "A" {}

	print(A().Test) -- nil
end)
```


### `__NoRawSet__`

Set the class's objects so save value to non-existent fields on them will be denied.

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	__NoRawSet__()
	class "A" {}

	o = A()
	o.age = 10 -- Error: The object can't accept field that named "age"
end)
```

Also you can mark a class so its object's fields can be assigned no matter how the platform settings is applied

```lua
PLOOP_PLATFORM_SETTINGS = { OBJECT_NO_RAWSEST = true, OBJECT_NO_NIL_ACCESS = true }

require "PLoop"

PLoop(function(_ENV)
	__NoRawSet__(false)
	class "A" {}

	A().Test = 123 -- Okay
end)
```

Even with the no rawset mode on, you still can override the object's method with functions:

```lua
PLOOP_PLATFORM_SETTINGS = { OBJECT_NO_RAWSEST = true, OBJECT_NO_NIL_ACCESS = true }

require "PLoop"

PLoop(function(_ENV)
	class "A" { Test = function() end }

	o = A()
	o.Test = 123    -- Error
	o.Test = print  -- Okay
	o.Test1= print  -- Error
end)
```


### `__ObjectAttr__`

Mark the class, so the attributes can be used on its objects.

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	class "__Reg__" (function(_ENV)
		extend "IAttachAttribute"

		function AttachAttribute(self, target, targettype)
			print("Attach To", target.name)
		end

		property "AttributeTarget"  { set = false, default = AttributeTargets.Object }
	end)

	__ObjectAttr__()
	class "A" { name = String }

	__Reg__()
	A{ name = "Ann" } -- Attach To Ann
end)
```


### `__ObjFuncAttr__`

Set the class's objects so functions that be assigned on them will be modified by the attribute system(target type is function)

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	__ObjFuncAttr__()
	class "A" {}

	o = A()

	__Async__()
	function o:Test()
		print(coroutine.running())
	end

	o:Test() -- thread: 02F195E8
end)
```

If the class is also marked with `__NoRawSet__`, we still can assign any functions on the object since the `__ObjFuncAttr__` means the object'll be used as a container of functions.


### `__ObjectSource__`

Set the class's objects to save the source where it's created

Attribute Targets:
* System.AttributeTargets.Class

Usage:

```lua
require "PLoop"

PLoop(function(_ENV)
	__ObjectSource__()
	class "A" {}

	o = A()

	print(Class.GetObjectSource(o)) -- @path_to_file\file.lua:7
end)
```


### `__Require__`

Set a require class to the target interface.

Attribute Targets:
* System.AttributeTargets.Interface


### `__Return__`

Bind return formats to the method or functions

Attribute Targets:
* System.AttributeTargets.Function
* System.AttributeTargets.Method

Usage:

If the validation is failed, since the call is finished, the system will try locate the source of the function:

```lua
require "PLoop"

PLoop(function(_ENV)
	__Return__{ String }
	function Test() return 1 end

	-- Error: path_to_file:5: The Test Return: System.String - the 1st return value must be string, got number
	Test()
end)
```

We also can bind several return formats to one function:

```lua
require "PLoop"

PLoop(function(_ENV)
    __Return__{ String } { Boolean }
    function Test() return 1 end

    -- lua: path_to_file:5: The Test should return:
    --  Return: System.String
    --  Return: System.Boolean
    Test()
end)
```

Like the `__Arguments__`, we can use optinal and varargs style like :

```lua
require "PLoop"

PLoop(function(_ENV)
    __Return__{ String/"hello", Number * 0 }
    function Test() return nil, 1, 2, 3 end

    -- hello    1   2   3
    print(Test())
end)
```

In most the case, we require the child classes that extend the interface should declare the methods with fixed type return values:

```lua
require "PLoop"

PLoop(function(_ENV)
    interface "IA" (function(_ENV)
        __Return__{ String, Number * 1 }:AsInheritable()
        __Abstract__() function Test() end
    end)

    class "A" { IA, Test = function() return "hi" end }

    -- Error: path_to_file:9: The A.Test Return: System.String, ... as System.Number - the ... must contains at least 1 return values
    A():Test()
end)
```


### `__Sealed__`

Seal the enum, struct, interface or class, so they can't be re-defined.

Attribute Targets:
* System.AttributeTargets.Enum
* System.AttributeTargets.Struct
* System.AttributeTargets.Interface
* System.AttributeTargets.Class


### `__Set__`

Set the set modifier to a property.

Attribute Targets:
* System.AttributeTargets.Property


### `__SingleVer__`

Set the class as a single version class, so all old objects will share the newest definition

Attribute Targets:
* System.AttributeTargets.Class


### `__Static__`

Set the object methods or object features as static, so they can only be used by the struct, interface or class itself.

Attribute Targets:
* System.AttributeTargets.Method
* System.AttributeTargets.Event
* System.AttributeTargets.Property


### `__Super__`

Set a super class to the target class

Attribute Targets:
* System.AttributeTargets.Class


### `__SuperObject__`

Whether the class's objects use the super object access style like `super[self]:Method()`, `super[self].Name = xxx`.

Attribute Targets:
* System.AttributeTargets.Class

Usage:
```lua
-- make sure class A use the super object access style
__SuperObject__(true)
class "A" {}

-- make sure class B don't use the super object access style
__SuperObject__(false)
class "B" {}
```


### `__Throwable__`

Mark the property as throwable.

Attribute Targets:
* System.AttributeTargets.Property
