# threading

The **System.Threading** contains several tools to easy the using of Lua's coroutines. Like thread pool, async declaration, iterator generator and lock mechanism(this will be discussed in the multi-os thread part).

## Table of Contents
* [ThreadPool](#threadpool)
	* [Run a function as coroutine](#run-a-function-as-coroutine)
	* [Run a function as an iterator](#run-a-function-as-an-iterator)
	* [Get the coroutine or wrap function](#get-the-coroutine-or-wrap-function)
* [`__Async__`](#`__async__`)
* [`__Iterator__`](#`__iterator__`)


## ThreadPool

The thread pool is used to genreate coroutines and recycle them automatically when their tasks are finished. Nomrally the Lua run on a single OS thread, so there is no need to create multi thread pools, we can use **System.ThreadPool.Default** for our tasks directly.

### Run a function as coroutine

We can process any function within a coroutine provided by the thread pool with **ThreadCall** method:

```lua
require "PLoop" (function(_ENV)
	function test(...)
		print(coroutine.running(), ...)
	end

	ThreadPool.Default:ThreadCall(test, 1, 2, 3)  -- thread: 00EE3C90	1	2	3
	ThreadPool.Default:ThreadCall(test, 1, 2, 3)  -- thread: 00EE3C90	1	2	3
end)
```

We can see the coroutine is re-used, in Lua, the creation of a coroutine is expensive, through the thread pool and special design of those coroutines we can save many cost.


### Run a function as an iterator

In many scenario, we should keep the context of an iterator, it's hard to be done with the common iterator generator since we should always keep upvalues to control the iterators. With the coroutine, we can yield values for many times, it's easy to use them as iterator:

```lua
require "PLoop" (function(_ENV)
	function Fibonacci(maxn)
		local n0, n1 = 1, 1

		coroutine.yield(0, n0)
		coroutine.yield(1, n1)

		local n = 2

		while n <= maxn  do
			n0, n1 = n1, n0 + n1
			coroutine.yield(n, n1)
			n = n + 1
		end
	end

	-- 0    1
	-- 1    1
	-- 2    2
	-- 3    3
	-- 4    5
	-- 5    8
	for k, v in ThreadPool.Default:GetIterator(Fibonacci, 5) do
		print(k, v)
	end

	-- we also can pass the argument later
	for k, v in ThreadPool.Default:GetIterator(Fibonacci), 5 do
		print(k, v)
	end
end)
```


### Get the coroutine or wrap function

In the previous examples, we let the system take controls of those coroutines. We also can get a coroutine or a wrap from the thread pool like the original coroutine.create or coroutine.wrap.

```lua
require "PLoop" (function(_ENV)
	function test(...)
		print(coroutine.running(), ...)
		coroutine.yield()
		return ...
	end

	local thread = ThreadPool.Default:GetThread(test)
	print(thread)                       -- thread: 00F8D4C0
	coroutine.resume(thread, 1, 2, 3)   -- thread: 00F8D4C0    1    2    3
	print(coroutine.resume(thread))     -- true    1    2    3

	local wrap = ThreadPool.Default:GetThread(test, true)
	print(wrap)                         -- function: 011F1058
	wrap(1, 2, 3)                       -- thread: 010C0E30    1    2    3
	print(wrap(1, 2, 3))                -- 1    2    3
end)
```

The coroutine used here are not recyclable, the different between it to the coroutine.create(wrap) is the thread pool itself is an argument to the thread, so the thread know which thread pool its generated throught the call stack, it's very useful in multi-os thread platforms, the system'll able to track the context of the thread.


## `__Async__`

It's not a good code style to keep using the **ThreadPool.Default**, and in multi-os thread platforms, we shouldn't use the default pool, we should use **ThreadPool.Current** to get a auto-created thread pool for current os-thread(we'll see more in Context chapter).

It'd be difficult for the user to choose the pool, so to simple the job, the **Threading** also provide two attribute for the job.

The `__Async__` attribute will use the **ThreadCall** method to process the target function.

Attribute Targets:
* System.AttributeTargets.Function
* System.AttributeTargets.Method

```lua
require "PLoop" (function(_ENV)
	__Async__()
	function test(...)
		print(coroutine.running(), ...)
	end

	test(1, 2, 3)  -- thread: 00F73208	1	2	3
	test(1, 2, 3)  -- thread: 00F73208	1	2	3
end)
```

Now, the calling style will be a simple function call.


## `__Iterator__`

The `__Iterator__` attribute will use the **GetIterator** method to process the target function.

Attribute Targets:
* System.AttributeTargets.Function
* System.AttributeTargets.Method

```lua
require "PLoop" (function(_ENV)
	__Iterator__()
	function Fibonacci(maxn)
		local n0, n1 = 1, 1

		coroutine.yield(0, n0)
		coroutine.yield(1, n1)

		local n = 2

		while n <= maxn  do
			n0, n1 = n1, n0 + n1
			coroutine.yield(n, n1)
			n = n + 1
		end
	end

	-- 0    1
	-- 1    1
	-- 2    2
	-- 3    3
	-- 4    5
	-- 5    8
	for k, v in Fibonacci(5) do
		print(k, v)
	end
end)
```

The iterators will works just like **pairs**.