# collections

When we need handle a group of simpliar objects or values, we could use a collection. In Lua, there are two ways to contain values. One is the indexed list, all keys are number and ordered, like `{1, 2, true, "test", 5}`, the other is hash tables of key-value pairs like `{ a = 1, b = 2}`.

The **System.Collections** is designed to provide support for collection operations.

## Table of Contents

* [System.Collections.Iterable](#systemcollectionsiterable)
* [System.Collections.IList](#systemcollectionsilist)
* [System.Collections.ICountable](#systemcollectionsicountable)
* [System.Collections.IIndexedList](#systemcollectionsiindexedlist)
* [System.Collections.List](#systemcollectionslist)
	* [The creation of List](#the-creation-of-list)
	* [The method of the List](#the-method-of-the-list)
	* [List As Template](#list-as-template)
* [System.Collections.XList](#systemcollectionsxlist)
* [System.Collections.Array](#systemcollectionsarray)
* [System.Collections.IDictionary](#systemcollectionsidictionary)
	* [The creation of the Dictionary](#the-creation-of-the-dictionary)
	* [The method of the Dictionary](#the-method-of-the-dictionary)
* [System.Collections.XDictionary](#systemcollectionsxdictionary)
* [System.Collections.Proxy](#systemcollectionsproxy)
* [Serialization](#serialization)


## System.Collections.Iterable

The interface is the root of the collection system, any collection types must extend it directly or indirectly, and must provide a **GetIterator** method used to return an iterator that can be used by the generic for.

```lua
require "PLoop" (function(_ENV)
	class "Array" (function(_ENV)
		extend "Iterable"

		GetIterator = ipairs
	end)

	v = Array{1, 2, 3, 4, 5, 6}

	for i, v in v:GetIterator() do
		print(i, v)
	end
end)
```

The **ipairs** and **pairs** can be used as the **GetIterator** method, they are also the method to the **List** and **Dictionary** class.

We could also create iterable types with `__Iterator__` attribute:

```lua
require "PLoop" (function(_ENV)
	class "Fibonacci" (function(_ENV)
		extend "IList"

		local yield = coroutine.yield

		__Iterator__()
		function GetIterator(self)
			local maxn = self.MAX
			local n0, n1 = 1, 1

			yield(0, n0)
			yield(1, n1)

			local n = 2

			while n <= maxn  do
				n0, n1 = n1, n0 + n1
				yield(n, n1)
				n = n + 1
			end
		end

		function __new(self, max)
			return { MAX = max }, true
		end
	end)

	-- 1,1,2,3,5,8,13,21,34,55,89
	print(Fibonacci(10):Join(","))
end)
```


## System.Collections.IList

In the previous example, the **IList** is an interface extend the **Iterable**, it contains several useful methods like the **Join**.

The **IList** interface represents the list collections that only elements has meanings, the key can be ignored by operations. The is a design goal, there is no rule to enforce it. For key-value paris in the list collection, the key don't need to be ordered index, the **IList** only care the value part. It provide the basic operations of the IList object.


Queue Method                             |Description
:----------------------------------------|:--------------------------------
Filter(self, func)                       |pass the list elements into the function, if the return value is non-false, the element would be used for next operations
Filter(self, name, value)                |If the `element[name] == value`, the element would be used for next operations
Map(self, func)                          |pass the list elements into the function and use the return value as new elements for next operations
Map(self, name)                          |use the `element[name]` as the new elements for next operations
Range(self[, start[, stop[, step]]])     |Only the elements in the range and fit to the step will be used for next operations, the *start*'s default value is 1, the *stop* is -1 and the *step* is 1

Final Method                             |Description
:----------------------------------------|:--------------------------------
All(self, func, ...)                     |pass the element with ... into the function, if the return value is false or nil, the final result is false, if all elements passed the function checking, the final result is true
Any(self, func, ...)                     |pass the element with ... into the function, if any return value is non-false, the final result is true, if all elements can't pass the function checing, the final result is false
Each(self, func, ...)                    |call the function with each elements and the ... argument
Each(self, name, ...)                    |if the `element[name]` is the element's method(function), the object method will be called with those ... argument, otherwise `element[name] = ...` will be used
First(self, func, ...)                   |pass the element with ... into the function, if the return value is non-false, return the current element
First(self)                              |return the first element if existed
Last(self, func, ...)                    |return the last element that match the func
Last(self)                               |return the last element
Reduce(self, func[, init])               |used to combine the elements, you can find the example in the above
ToList(self[, listtype])                 |save the elements into a new list type object, the default listtype is the **List**
Sum(self)                                |calculate the sum of the list, to reduce the cost, there is no validation in the method to make sure all elements are numbers(just use the tonumber to convert them).
Join(self[, sep])                        |get the concatenation of the List, to reduce the cost, there is no validation in the method to make sure all elements are string.

**All the *func* for those methods are callable value like function, lambda expression or object with `__call` meta-method.**

There are two types of those method: the **queue** method like the **Range**, **Filter** and **Map**, the **final** method like **Each**, **Reduce** and others.

The queue method is used to queue operations with options, and when the final method is called, the queued operations'll be converted to a whole iterator so the final method can traverse the result to do the jobs. Those queue operations'll be done by a **System.Collections.ListStreamWorker** object, so the operation can be processed like:

```lua
require "PLoop"

PLoop(function(_ENV)
	-- [Operation] List(10):Range(1, -1, 2):Map("x=>x^2"):Each(print)

	local obj = List(10)

	-- get a stream worker for next operations
	local worker = obj:Range(1, -1, 2)

	-- the same worker
	worker = worker:Map("x=>x^2")

	-- the final method
	for _, v in worker:GetIterator() do
		print(v)
	end
end)
```

Since the stream workers are inner objects that don't need be controlled by the users, the system can recycle them and re-use them for the next stream operations. So there is no need to care about the inner operations and you shouldn't take the stream worker out for other usage.

The stream workers use the `__Iterator__` attribute for its **GetIterator** method, so all the traversing are done in a recyclable coroutine, there is no cache or anonymous method created during those operations.


## System.Collections.ICountable

The **ICountable** interface extend the **IList**, and provide an abstract property "Count" used to get the object's element count, so the system don't need to traverse the IList object to get the element count. The default is using the `#self`.

For an advance example:

```lua
require "PLoop" (function(_ENV)
	class "Fibonacci" (function(_ENV)
		extend "ICountable"

		property "Count" { set = false, field = "MAX" }

		__Iterator__()
		function GetIterator(self)
			-- skip
		end

		function __new(self, max)
			return { MAX = max }, true
		end
	end)

	-- 10
	print(Fibonacci(10).Count)
end)
```

It's very important to extend the **ICountable** interface if possible, many operations the systme provided will check it, otherwise, the system can only get the count by traverse the object.


## System.Collections.IIndexedList

The **IIndexedList** interface represents the indexed list collections that can use obj[idx] to access the its elements. It also extend the **ICountable** interface.

It's very important to extend the **IIndexedList** interface if possible, many operations the systme provided will check it, otherwise, the system can only get the element by traverse the object.

The **IIndexedList** interface also provided several sort methods:


Sort                                            |Description
:-----------------------------------------------|:--------------------------------
Reverse(self[, start[, stop]])                  |Reverse the indexed list, the start's default is 1, the stop is -1, it's a *sort* method since it'd change the list itself
BubbleSort(self, [compare[, start[, stop]]])    |Use the bubble sort on the indexed list, the default compare is function(x, y) return x < y end
CombSort(self, [compare[, start[, stop]]])      |Use the comb sort on the indexed list
HeapSort(self, [compare[, start[, stop]]])      |Use the heap sort on the indexed list
InsertionSort(self, [compare[, start[, stop]]]) |Use the insertion sort on the indexed list
MergeSort(self, [compare[, start[, stop]]])     |Use the merge sort on the indexed list
QuickSort(self, [compare[, start[, stop]]])     |Use the quick sort on the indexed list
SelectionSort(self, [compare[, start[, stop]]]) |Use the selection sort on the indexed list
Sort(self, [compare[, start[, stop]]])          |Use the lua's table.sort on the indexed list
TimSort(self, [compare[, start[, stop]]])       |Use the tim sort on the indexed list


**All the *compare* for those methods are callable value like function, lambda expression or object with `__call` meta-method, its default is `x, y => x < y`**.

Here is a test:

```lua
require "PLoop" (function(_ENV)
	local random = math.random
	local function val() return random(500000) end

	function test(cnt, sortMethod)
		collectgarbage()

		local st = os.clock()

		for i = 1, cnt do
			local lst = List(1000, val)
			lst[sortMethod](lst)
		end

		print(sortMethod, "Cost", os.clock() - st)
	end

	test(100, "BubbleSort")
	test(100, "CombSort")
	test(100, "HeapSort")
	test(100, "InsertionSort")
	test(100, "MergeSort")
	test(100, "QuickSort")
	test(100, "SelectionSort")
	test(100, "Sort")
	test(100, "TimSort")
end)
```

The test result in Lua 5.1 is :

```
BubbleSort      Cost    13.916
CombSort        Cost    0.422
HeapSort        Cost    0.484
InsertionSort   Cost    4.772
MergeSort       Cost    0.535
QuickSort       Cost    0.281
SelectionSort   Cost    6.417
Sort            Cost    0.109
TimSort         Cost    0.547
```

So the table.sort is still the best choice for Lua 5.1.

For the luajit:

```
BubbleSort      Cost    0.269
CombSort        Cost    0.033
HeapSort        Cost    0.038
InsertionSort   Cost    0.125
MergeSort       Cost    0.046
QuickSort       Cost    0.018
SelectionSort   Cost    0.075
Sort            Cost    0.084
TimSort         Cost    0.036
```


## System.Collections.List

The **List** class extend the **IIndexedList** interface, so it has all the features provided by the **IList**, **ICountable** and **IIndexedList**. It's also one of the main collection class.


### The creation of List

There are several ways to create a **List** object:

Constructor                        |Result
:----------------------------------|:--------------------------------
List(table)                        |Convert the input table as a list object, no new table would be generated
List(IList)                        |Copy all elements form the other list object(may be other list type's object)
List(iterator[, object[, index]])  Use it like List(ipairs{1, 2, 3})，use the result of the iterator as list elements
List(count, func)                  |Repeat func(i) for _count_ times, use those result as list elements
List(count, init)                  |Create a list with _count_ elements, all values are the init
List(...)                          |Use the arguments as the list elements, it's the same like List{...}

Here are examples:

```lua
require "PLoop" (function(_ENV)
	-- List(table)
	o = {1, 2, 3}
	print(o == List(o))  -- true

	-- List(count)
	v = List(10)         -- {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}

	-- List(count, func)
	v = List(10, function(i) return math.random(100) end) -- {46, 41, 85, 80, 62, 37, 29, 91, 62, 37}

	-- List(...)
	v = List(1, 5, 4)    -- {1, 5, 4}
end)
```


### The method of the List

The **List** class has provide basic method for list operations:

Method                                   |Description
:----------------------------------------|:--------------------------------
Clear(self)                              |Clear the list
Contains(self, item)                     |Whether the item existed in the list
GetIterator(self)                        |Return an interator for traverse, it's the iparis
IndexOf(self, item)                      |Get the index of an item if it existed in the list
Insert(self[, index], item)              |Insert an item in the list
Remove(self, item)                       |Remove an item from the list
RemoveByIndex(self[, index])             |Remove and item by index or from the tail if index not existed
Extend(self, table)                      |Extend the list with another index table
Extend(self, listobject)                 |Extend the list with another IList object
Extend(self, iterator[, object[, index]])|Extend the list with an iterator

```lua
require "PLoop"

PLoop(function(_ENV)
	obj = List(10)

	print(obj:Remove()) -- 10
end)
```

```lua
require "PLoop" (function(_ENV)
	-- 1,2,3,4,5,6,7,8,9
	print(List{1, 2, 3, 4}:Extend{5, 6, 7, 8, 9}:Join(","))
end)
```


### List As Template

The **List** is a template class, we can give it any type, so it'll do the type validation for several method:

```lua
require "PLoop" (function(_ENV)
	l = List[Number]()

	-- Error: the calling style must be one of the follow:
	--    Usage: Anonymous:Insert(System.Integer, System.Number)
	--    Usage: Anonymous:Insert(System.Number)
	l:Insert("hi")
end)
```

BTW. the assignment like `l[1] = "hi"` wouldn't be checked, since override an existing value can't trigger the `__newindex`, so there is no need to do.


## System.Collections.XList

If we want do some tasks by traverse from 1 to 10000, use `List(10000)` would be a greater waste since we don't really need to save 10000 numbers in a list.

The **XList** is a dynamic list, it don't save any elements in itself, we can construct it with

Constructor                        |Result
:----------------------------------|:--------------------------------
XList(start, stop[, step])         |Traverse from start to stop with step
XList(stop)                        |Traverse from 1 to stop with 1 as step
XList(IList)                       |Traverse the IList object
XList(table)                       |Traverse the lua table
XList(iterator[, object[, index]]) |Traverse through the iterator

Here are examples:

```lua
require "PLoop"

PLoop (function(_ENV)
	-- start, stop[, step]
	XList(5, 10, 2):Each(print)

	-- step can be negative number
	XList(4, 1, -1):Each(print)

	-- stop
	-- start and step will be 1 as default
	XList(10):Map("x=>x^2"):Each(print)

	-- iter[, lst[, index]]
	XList(ipairs{1, 2, 3, 4}):Each(print)

	XList{1, 2, 3, 4}:Each(print)

	-- The stream worker is also a list object,
	-- but the stream worker can only be traversed by once
	-- then it'd be recycled
	XList(List(10):Range(5, 8)):Each(print)
end)
```

It'll avoid the creation of bigger tables, it's a class that extend the **System.Collections.IList**, so you always can use the Map to create the real values.


## System.Collections.Array

The **Array** is directly inherit the **List**, it can provide event handlers to all elements at once, and for the new elements added lately.

```lua
require "PLoop" (function(_ENV)
	class "Person" (function(_ENV)
		event "OnNameChanged"

		property "Name" { type = String, event = OnNameChanged }
	end)

	arPerson = Array[Person]()

	function arPerson:OnNameChanged(new)
		print("new name is " .. new)
	end

	ann = Person()
	arPerson:Insert(ann)

	-- new name is Ann
	ann.Name = "Ann"
end)
```

You must keep using **Insert** method to make sure the event handlers can registered to the new elements.


## System.Collections.IDictionary

The **IDictionary** interface is also extend the **Iterable** interface, it represents the key-value pairs collections, unlike the **IList**, the list system only care the values, the dictionary system take care about all the keys and values.

It provide several features for dictionary operations:

Queued Method                            |Description
:----------------------------------------|:--------------------------------
Filter(self, func)                       |Pass the key-value pair into the function, if the return value is non-false, the key-value pair should be used in next operations
Map(self, func)                          |Pass the key-value pair into the function, use the return value as new value with the key into next operations


Final Method                             |Description
:----------------------------------------|:--------------------------------
Each(self, func, ...)                    |Pass each key-value pairs into the function
GetKeys(self)                            |Return an iterator used like `for index, key in dict:GetKeys() do print(key) end`
GetValues(self)                          |Return an iterator used like `for index, value in deict:GetValues() do print(value) end`
Reduce(self, func, init)                 |Combie the key-value pairs, we'll see an example later
ToDict(self, class)                      |Create a new dictionary object contains the key-value pairs from the object, default use Dictionary


Property                                 |Description
:----------------------------------------|:--------------------------------
Keys                                     |Get a list stream worker of keys
Values                                   |Get a list stream worker of the values


Like the list system, the queued method also would return a recyclable **DictionaryStreamWorker** object, queue the operations and do the traverse for the final method.

With the **Keys** and **Values** property, it'd return a list stream worker so we can use all list operations on the keys or values.

Here are examples:

```lua
require "PLoop"

PLoop(function(_ENV)
	-- print all keys in the _G with order
	Dictionary(_G).Keys:ToList():Sort():Each(print)

	-- Calculate the sum of the values
	print(Dictionary{ A = 1, B = 2, C = 3}:Reduce(function(k, v, init) return init + v end, 0))
end)
```


### The creation of the Dictionary

Like the **List**, we also have several ways to create a **Dictionary** object:

Constructor                          |Result
:------------------------------------|:--------------------------------
Dictionary()                         |Create an empty dictionary object
Dictionary(table)                    |Convert the input table as a dictionary object
Dictionary(table, table)             |the two array tables or lists, the first table's elements would be used as keys and the second table's element would be used as values to create the dictionary
Dictionary(dictionary)               |Copy other dictionary's key-value pairs to create a new dictionary
Dictionary(iter, obj, index)         |Use key, value pairs generated by the iterator to create a new dictionary

Here is some examples:

```lua
require "PLoop" (function(_ENV)
	Dictionary(_G) -- Convert the _G to a dictionary

	-- key map to key^2
	lst = List(10)
	Dictionary(lst, lst:Map("x=>x^2")):Each(print)
end)
```


### The method of the Dictionary

The dictionary are normally hash tables, you can use **pairs** to traverse them or use `obj[key] = value` to modify them, the **Dictionary**'s **GetIterator** method is just the **pairs**.

There are only one method decalred in the Dictionary class:

Method                                   |Description
:----------------------------------------|:--------------------------------
Update(self, table)                      |Update the dictionary with the table's value
Update(self, IDictionary)                |Update the dictionary with the IDictionary object's value
Update(self, iter[, obj[, idx])          |Update the dictionary with the iterator's key-value pairs

```lua
require "PLoop" (function(_ENV)
	v = Dictionary(List(5), List(5))
	v:Update{ [3] = 9, [4] = 16 }

	-- 1   1
	-- 2   2
	-- 3   9
	-- 4   16
	-- 5   5
	v:Each(print)
end)
```

## System.Collections.XDictionary

Like the dynamic List, we also can create **XDictionary** objects based on other dictionary, iterator, key-value lists.

Constructor                          |Result
:------------------------------------|:--------------------------------
XDictionary(table)                   |Traverse the target table
XDictionary(table, table)            |the two array tables or lists, the first table's elements would be used as keys and the second table's element would be used as values
XDictionary(dictionary)              |Traverse other dictionary's key-value pairs
XDictionary(iter, obj, index)        |Traverse through the iterator


Since it won't generate key-value pairs based on those inputs, it'd save many time cost for several conditions:

```lua
require "PLoop"

PLoop(function(_ENV)
	-- key value lists
	XDictionary(List(4), {"1st", "2nd", "3rd", "4th"}):Each(print)

	-- hash table or dictionary
	XDictionary(_G):Each(print)

	-- iterator
	XDictionary(pairs(_G)):Each(print)
end)
```

## System.Collections.Proxy

Since the **Dictionary** has several methods and two properties, those names can't be used as the element key, so if you can't make sure it'd be not a problem, you'd better use the **Proxy** instead of the **Dictionary**.

The **Proxy** works like the **Dictionary**, but it provide an indexer property *Item* to set/get the elements:

```lua
require "PLoop" (function(_ENV)
	p = Proxy(_G)

	-- nil
	print(getmetatable(_G))

	p.Item["A"] = 123

	-- true
	print(p.Item["A"] == _G.A)

	p.A = 456

	-- 456
	print(p.Item.A)
end)
```

You also can use `p[key] = element` if it won't cause conflict with the method or property, and unlick the **Dictionary** the target table won't be used as a **Proxy** object.


## Serialization

To simplify the object's serialization, the **List** and **Dictionary** are designed to be serializable:

```lua
require "PLoop" (function(_ENV)
	import "System.Serialization"

	o = Dictionary{ A = Date(2013, 8, 13), B = Date(2017, 4, 2), C = Date(2018, 3, 14)}

	v = Serialization.Serialize( StringFormatProvider{ Indent = true, ObjectTypeIgnored = false  }, o )

	--	{
	--		__PLoop_Serial_ObjectType = "System.Collections.Dictionary",
	--		[1] = {
	--			__PLoop_Serial_ObjectType = "System.Collections.List",
	--			[1] = "A",
	--			[2] = "C",
	--			[3] = "B"
	--		},
	--		[2] = {
	--			__PLoop_Serial_ObjectType = "System.Collections.List",
	--			[1] = {
	--				__PLoop_Serial_ObjectType = "System.Date",
	--				time = 1376366400
	--			},
	--			[2] = {
	--				__PLoop_Serial_ObjectType = "System.Date",
	--				time = 1521000000
	--			},
	--			[3] = {
	--				__PLoop_Serial_ObjectType = "System.Date",
	--				time = 1491105600
	--			}
	--		}
	--	}
	print(v)

	o = Serialization.Deserialize( StringFormatProvider(), v)

	-- 	A	2013-08-13 12:00:00
	-- 	B	2017-04-02 12:00:00
	-- 	C	2018-03-14 12:00:00
	o:Each(print)
end)
```

The **System.Date** is a serializable type, so it can be used for the serialization.

Take a look at the *v*, the dictionary object is splitted to two **List** objects, one for keys, one for values. We have those `__PLoop_Serial_ObjectType` to help use deserialize the result, but in many cases, we can't keep the object type into the result:

```lua
require "PLoop" (function(_ENV)
	import "System.Serialization"

	o = Dictionary{ A = Date(2013, 8, 13), B = Date(2017, 4, 2), C = Date(2018, 3, 14)}

	v = Serialization.Serialize( StringFormatProvider{ Indent = true, ObjectTypeIgnored = true  }, o )

	--	{
	--		[1] = {
	--			[1] = "A",
	--			[2] = "C",
	--			[3] = "B"
	--		},
	--		[2] = {
	--			[1] = {
	--				time = 1376366400
	--			},
	--			[2] = {
	--				time = 1521000000
	--			},
	--			[3] = {
	--				time = 1491105600
	--			}
	--		}
	--	}
	print(v)

	o = Serialization.Deserialize( StringFormatProvider(), v, Dictionary)

	--	A	table: 03026B90
	--	C	table: 03026CA8
	--	B	table: 03026CF8
	o:Each(print)
end)
```

Although we pass the **Dictionary** as the type, we can't pass the key and value's type to the serialization system.

To solve this problem, the **List** and **Dictionary** are designed as [Template class](#template-class), so we can add the key and value type to the **Dictionary**:

```lua
require "PLoop" (function(_ENV)
	import "System.Serialization"

	o = Dictionary{ A = Date(2013, 8, 13), B = Date(2017, 4, 2), C = Date(2018, 3, 14)}

	v = Serialization.Serialize( StringFormatProvider{ Indent = true, ObjectTypeIgnored = true  }, o )

	--	{
	--		[1] = {
	--			[1] = "A",
	--			[2] = "C",
	--			[3] = "B"
	--		},
	--		[2] = {
	--			[1] = {
	--				time = 1376366400
	--			},
	--			[2] = {
	--				time = 1521000000
	--			},
	--			[3] = {
	--				time = 1491105600
	--			}
	--		}
	--	}
	print(v)

	o = Serialization.Deserialize( StringFormatProvider(), v, Dictionary[{String, Date}])

	--	A	2013-08-13 12:00:00
	--	B	2017-04-02 12:00:00
	--	C	2018-03-14 12:00:00
	o:Each(print)
end)
```

Now, the serialization system can know the key and value's types and we can have our objects back now.