# Tools

There are several useful apis or types provided by the **PLoop**, you can use them or simply replaced them if you like.

## Table of Contents

* [System.Toolset](#systemtoolset)
	* [wipe](#wipe)
	* [safeset](#safeset)
	* [clone](#clone)
	* [loadsnippet](#loadsnippet)
	* [parseindex](#parseindex)
	* [validateflags](#validateflags)
	* [turnonflags](#turnonflags)
	* [turnoffflags](#turnoffflags)
* [System.ICloneable](#systemicloneable)
* [System.Date](#systemdate)
	* [The creation of Date](#the-creation-of-date)
	* [Static Property of Date](#static-property-of-date)
	* [Static Method of Date](#static-method-of-date)
	* [Property of Date](#property-of-date)
	* [Method of Date](#method-of-date)
* [System.Logger](#systemlogger)
	* [Log Level](#log-level)
	* [Log Handler](#log-handler)
	* [Log Message](#log-message)
		* [Time Format](#time-format)
		* [Prefix](#prefix)
		* [Format Message](#format-message)
	* [Log Function](#log-function)
	* [Default Log](#default-log)
* [System.Recycle](#systemrecycle)
	* [Recycle Event](#recycle-event)
	* [Recycle Object with Type](#recycle-object-with-type)


## System.Toolset

To build the prototype and the whole type system, the **Prototype.lua** has created several useful methods and can be shared through the **System.Toolset** interface, you can also assign method into the interface if you need.


### wipe

Wipe and return the table

* Params:
	* target        - the table should be wiped
* Return:
	* target      	- the target table


### safeset

Safe save the value into a table, in single os thread platform, it works like rawset and return the target table directly, in multi os thread platform, if the key don't existed in the table, a new table'll be created save the new key-value pair and copy all data from the old one, then the new table'd be returned, this is used to resolve the re-hash problem when multi os threads access the table at the same time.

* Params:
	* target        - the target table
	* key           - the key
	* value         - the value
* Return:
	* table         - the target table or a new one to replace the target


### clone

clone the value or object

* Params:
	* value         - the value should be cloned
	* deep          - whether use deep clone
	* safe          - whether the value has recursive reference
* Return:
	* clone         - the cloned value


### loadsnippet

load the snippet and return the result

* Params:
	* chunk         - the code
	* souorce       - the source name
	* env           - the environment to the code, default `_G`
* Return:
	* result        - the load result
	* error         - the error message if failed to load


### parseindex

Convert an index number to string like '1st', '2nd', '7th'

* Params:
	* index         - the index
* Return:
	* result        - the converted result


### validateflags

validate flags values

* Params:
	* chkvalue      - the check value, must be 2^n
	* targetvalue   - the target value
* Return:
	* boolean      	- true if the target value contains the chkvalue


### turnonflags

add the check value to the target value

* Params:
	* chkvalue      - the check value, must be 2^n
	* targetvalue   - the target value
* Return:
	* targetvalue   - the target value that contains the check value

### turnoffflags

remove the check value to the target value

* Params:
	* chkvalue      - the check value, must be 2^n
	* targetvalue   - the target value
* Return:
	* targetvalue   - the target value don't contains the check value


## System.ICloneable

For the *clone* method, here we take an example:

```lua
require "PLoop" (function(_ENV)
	class "A" { Name = String }

	data = { { }, A{ Name = "Ann" } }

	-- deep clone
	clone = Toolset.clone(data, true)

	-- true
	print(clone[1] ~= data[1])

	-- true
	print(clone[2] == data[2])
end)
```

With the deep clone, the `clone[1]` is a new table cloned from `data[1]`, but the data[2] is not cloned, since it's an object of class `A`, and we consider any object as a standalone value.

If we need the object be cloned, the object class must extend **System.ICloneable** and provide a **Clone** method that return a new copy of it:

```lua
require "PLoop" (function(_ENV)
	class "A" { ICloneable,
		Name    = String,
		Clone   = function(self)
			return A{ Name = self.Name }
		end
	}

	data = { { }, A{ Name = "Ann" } }

	-- deep clone
	clone = Toolset.clone(data, true)

	-- true
	print(clone[1] ~= data[1])

	-- false
	print(clone[2] == data[2])
end)
```

The **clone** API don't check **ICloneable**, it'll get the prototype of the object, here is the class *A*, and then get the prototype of the *A*, it's the **class** prototype, then try to retrieve the **Clone** method from the prototype, use it on the object if existed.

The **class.Clone** will check if the object extend **ICloneable**.


## System.Date

The **Date** object represents the date objects, they are cloneable and serializable.


### The creation of Date

Constructor                      |Arguments                                                                                      |Description
:--------------------------------|:----------------------------------------------------------------------------------------------|:-----------------
Date(time)                       |time:Integer/nil                                                                               |Create a date represents the time or now
Date(y, m, d, h, min, sec, utc)  |y:Integer, m:Integer, d:Integer, h:Integer/12, min:Integer/0, sec:Integer/0, utc:Boolean/false |Create a date represents the date settings

The **Date** class provided `__tostring`, `__lt`, `__eq`, `__sub` meta-methods, so we can directly change them to string, compare the data objects, and get there diff time through subtract operation:

```lua
require "PLoop" (function(_ENV)
	date = Date(2000, 10, 1)
	date2 = Date(date.time + 1000)

	print(date > date2) -- false
	print(date2)        -- 2000-10-01 12:16:40
	print(date2 - date) -- 1000
end)
```

### Static Property of Date

Static Proprety         |Description
:-----------------------|:-----------------------
Now                     |Gets a DateTime object that is set to the current date and time on this computer, expressed as the local time
GetTimeOfDay            |Gets and Sets the function that return the current time value(the total second from 1970/1/1 00:00:00)

```lua
require "PLoop" (function(_ENV)
	print(Date.Now)  -- 2019-03-10 11:02:39
end)
```

The **Date.GetTimeOfDay** normally is the *os.time*, since it's a system call, you can replace it like

```lua
require "PLoop" (function(_ENV)
	System.Date.GetTimeOfDay = ngx.time
end)
```

This is an example from [NgxLua][] works for [Openresty][].


### Static Method of Date

Static Method           |Arguments                                                               |Description
:-----------------------|:-----------------------------------------------------------------------|:-----------------------
Parse                   |String: NEString, Format:TimeFormat/"%Y-%m-%d %X", IsUTC:Boolean/false  |Parse a string data to Date object

The **TimeFormat** is a string combined with patterns, here is a full list:

Pattern |Description                                                                  |Result
:-------|:----------------------------------------------------------------------------|:-----------------------
%a      |Abbreviated weekday name                                                     |Thu
%A      |Full weekday namespace                                                       |Thursday
%b      |Abbreviated month name                                                       |Aug
%B      |Full month name                                                              |August
%c      |Date and time representation                                                 |Thu Aug 23 14:55:02 2001
%d      |Day of the month, zero-padded (01-31)                                        |23
%H      |Hour in 24h format (00-23)                                                   |14
%I      |Hour in 12h format (01-12)                                                   |02
%j      |Day of the year (001-366)                                                    |235
%m      |Month as a decimal number (01-12)                                            |08
%M      |Minute (00-59)                                                               |55
%p      |AM or PM designation                                                         |PM
%S      |Second (00-61)                                                               |02
%U      |Week number with the first Sunday as the first day of week one (00-53)       |33
%w      |Weekday as a decimal number with Sunday as 0 (0-6)                           |4
%W      |Week number with the first Monday as the first day of week one (00-53)       |34
%x      |Date representation                                                          |08/23/01
%X      |Time representation                                                          |14:55:02
%y      |Year, last two digits (00-99)                                                |01
%Y      |Year                                                                         |2001
%Z      |Timezone name or abbreviation

So the default time format for the parse would match date string like `2019-03-10 11:02:39`, it's also the default time format to convert a date object to string. Oonly several patterns can be used in the **Parse**:

Pattern |Description                                                                  |Result
:-------|:----------------------------------------------------------------------------|:-----------------------
%d      |Day of the month, zero-padded (01-31)                                        |23
%H      |Hour in 24h format (00-23)                                                   |14
%m      |Month as a decimal number (01-12)                                            |08
%M      |Minute (00-59)                                                               |55
%S      |Second (00-61)                                                               |02
%X      |Time representation                                                          |14:55:02
%Y      |Year                                                                         |2001

```lua
require "PLoop" (function(_ENV)
	print(Date.Parse("2010-01-02 12:31:45").Time)  -- 1262406705

	print(Date.Parse("2010/01/02 12:00:00", "%Y/%m/%d %X").Time) -- 1262404800
end)
```

### Property of Date

Proprety                |Description
:-----------------------|:-----------------------
Year                    |Gets or sets the year of the date
Month                   |Gets or sets the month of the date, 1-12
Day                     |Gets or sets the day of the date, 1-31
Hour                    |Gets or sets the hour of the date, 0-23
Minute                  |Gets or sets the minute of the date, 0-59
Second                  |Gets or sets the second of the date, 0-59
DayOfWeek               |Gets the day of the week, sunday is 1
DayOfYear               |Gets the day of the year
IsDaylightSavingTime    |Whether this instance of DateTime is within the daylight saving time range for the current time zone
Time                    |Gets the time that represent the date and time of this instance

You can assign any integer value to those values, the system will convert them:

```lua
require "PLoop" (function(_ENV)
	date     = Date(2000, 10, 1)
	date.Day = date.Day - 100
	print(date)  -- 2000-06-23 12:00:00
end)
```

### Method of Date

Method                  |Arguments                        |Description
:-----------------------|:--------------------------------|:-----------------------
Diff                    |date: Date                       |Return the diff second for the two Date object
ToString                |format:TimeFormat/"%Y-%m-%d %X"  |Converts the value of the current DateTime object to its equivalent string representation using the specified format.
ToUTCString             |format:TimeFormat/"!%Y-%m-%d %X" |Converts the value of the current DateTime object to its equivalent UTC string representation using the specified format.
AddYears                |years: Integer                   |Adds the specified number of years to the value of this instance, and return a new date object
AddMonths               |years: Integer                   |Adds the specified number of months to the value of this instance, and return a new date object
AddDays                 |years: Integer                   |Adds the specified number of days to the value of this instance, and return a new date object
AddHours                |years: Integer                   |Adds the specified number of hours to the value of this instance, and return a new date object
AddMinutes              |years: Integer                   |Adds the specified number of minutes to the value of this instance, and return a new date object
AddSeconds              |years: Integer                   |Adds the specified number of seconds to the value of this instance, and return a new date object
Clone                   |                                 |Return a Clone of the date oject

```lua
require "PLoop" (function(_ENV)
	date = Date(2000, 10, 1)
	print(date:AddSeconds(100) - date)  -- 100
end)
```


## System.Logger

The **Logger** is used to distribute log message based on log levels. Start with an example:

```lua
require "PLoop" (function(_ENV)
	log = Logger()

	log:AddHandler(print)

	-- [03/10/19 16:21:37]Save data is failed
	log(Logger.LogLevel.Warn, "%s is failed", "Save data")
end)
```

There are four parts of the logger system.

### Log Level

Each error message has a log level, the message could be discarded if its log level is lower than the logger's LogLevel property.

The enum **System.Logger.LogLevel** has defined the log levels:

* Trace  - 1, logs that contain the most detailed messages. These messages may contain sensitive application data. These messages should never be enabled in a production environment.
* Debug  - 2, logs that are used for interactive investigation during development. These logs should primarily contain information useful for debugging and have no long-term value.
* Info   - 3, logs that track the general flow of the application. These logs should have long-term value. This is also the default log level of Logger objects.
* Warn   - 4, logs that highlight an abnormal or unexpected event in the application flow, but do not otherwise cause the application execution to stop.
* Error  - 5, logs that highlight when the current flow of execution is stopped due to a failure. These should indicate a failure in the current activity, not an application-wide failure.
* Fatal  - 6, logs that describe an unrecoverable application or system crash, or a catastrophic failure that requires immediate attention.

We can change the logger's log level by property:

Proprety                |Description
:-----------------------|:-----------------------
LogLevel                |The log level

```lua
require "PLoop" (function(_ENV)
	log = Logger()

	log:AddHandler(print)

	-- [03/10/19 16:21:37]Save data is failed
	log(Logger.LogLevel.Warn, "%s is failed", "Save data")

	log.LogLevel = Logger.LogLevel.Error

	-- nothing output
	log(Logger.LogLevel.Warn, "%s is failed", "Save data")
end)
```


### Log Handler

The logger objects are used to distribute log messages, they need log handlers to print out the message or save to files.

A handler can be registered for all messages of all log levels, also can be registered for messages of special log level.

Method                  |Arguments                           |Description
:-----------------------|:-----------------------------------|:-----------------------
AddHandler              |handler:Callable, lvl:LogLevel/nil  |Add a handler to receive the log messages
RemoveHandler           |handler:Callable                    |Remove a log handler

```lua
require "PLoop" (function(_ENV)
	log = Logger()

	log:AddHandler(print, Logger.LogLevel.Warn)

	-- [03/10/19 16:21:37]Save data is failed
	log(Logger.LogLevel.Warn, "%s is failed", "Save data")

	-- nothing output
	log(Logger.LogLevel.Error, "%s is success", "Save data")
end)
```


### Log Message

A log message is combined with three part:


#### Time Format

In previous examples, there always a time mark at the beginning of the log message, through properties we can modify the time format or disable it:

Proprety                |Description
:-----------------------|:-----------------------
TimeFormat              |The time format, see **System.Date** for more details. Default `[%c]`
UseTimeFormat           |Whether use the time format. Default true


#### Prefix

We can set prefixes for each log level, the prefix will be added before the log message.

Method                  |Arguments                           |Description
:-----------------------|:-----------------------------------|:-----------------------
SetPrefix               |lvl:LogLevel, prefix:String/nil     |Set a prefix for a log level, thre prefix will be added to the message when the message is with the same log level

```lua
require "PLoop" (function(_ENV)
	log = Logger()

	log:AddHandler(print)
	log:SetPrefix(Logger.LogLevel.Warn, "[WARN]")

	-- [03/10/19 20:03:52][WARN]Save data is failed
	log(Logger.LogLevel.Warn, "%s is failed", "Save data")
end)
```


#### Format Message

The log messages will always contains several variables for informations, so we normally use *string.format* to generate the log message, but if the log message's log level is lower than the logger object's log level, it wouldn't be distributed, so there is no need to call the *string.format* to generate a new string.

The logger's message can be a format string, so the logger will format the string with data if it'd be distributed.


### Log Function

It's diffcult to always write down the log level, so there is an alternative way, we can get several log function from a logger with log level, those log function will be used to send the message with its log level:

```lua
require "PLoop" (function(_ENV)
	log = Logger()

	log:AddHandler(print)

	Warn = log[Logger.LogLevel.Warn]
	Error = log[Logger.LogLevel.Error]

	-- [03/10/19 20:22:57]Save data is failed
	Warn("%s is failed", "Save data")

	-- [03/10/19 20:22:57]File load failed
	Error("%s load failed", "File")
end)
```


### Default Log

The **Logger.Default** has provide a default logger used by the **PLoop** itself, almost all logs generated by **PLoop** is send to the default logger. You also could use it on your case:

```lua
require "PLoop" (function(_ENV)
	Info   = Logger.Default[Logger.LogLevel.Info]
	Warn   = Logger.Default[Logger.LogLevel.Warn]
	Error  = Logger.Default[Logger.LogLevel.Error]

	Logger.Default:AddHandler(print)

	-- [03/12/19 11:36:51][Warn]Something something not well
	Warn("Something something not well")
end)

```


## System.Recycle

**Recycle** is used as an object factory and recycle manager. For a simple example, we use it as to generate caches:

```lua
require "PLoop" (function(_ENV)
	ry = Recycle()

	t = ry()
	print(t)  -- table: 00DB3E00

	t2 = ry()
	print(t2) -- table: 00DB3E78

	ry(t)
	t3 = ry()
	print(t3) -- table: 00DB3E00
end)
```

So we can use `ry()` to pop a cache from the recycle object, and `ry(t)` to push back a cache so it'd be re-used for another purpose.

We also can use methods instead call the object directly:

```lua
require "PLoop" (function(_ENV)
	ry = Recycle()

	t = ry:Pop()
	print(t)  -- table: 00DB3E00

	t2 = ry:Pop()
	print(t2) -- table: 00DB3E78

	ry:Push(t)
	t3 = ry:Pop()
	print(t3) -- table: 00DB3E00
end)
```


### Recycle Event

There are three event of the **Recyle** object:

Proprety                |Param                   |Description
:-----------------------|:-----------------------|:-----------------------
OnInit                  |object                  |Fired when an object is created
OnPop                   |object                  |Fired when an un-used object is send out
OnPush                  |object                  |Fired when an no longer used object is put in

With these events, we can init the object, clear object when not used:

```lua
require "PLoop" (function(_ENV)
	ry = Recycle()

	function ry:OnPush(obj)
		Toolset.wipe(obj)
	end

	t = ry()
	t[1] = 1
	print(t, t[1])  -- table: 00E3CE58  1

	ry(t)

	t = ry()
	print(t, t[1])  -- table: 00E3CE58  nil
end)
```


### Recycle Object with Type

We can override the **New** method of a recycle object to provide the support of type, the method should be called when a new object need to be created:

```lua
require "PLoop" (function(_ENV)
	class "Person" { Name = { type = String, default = "Anonymous" } }

	ry = Recycle()

	function ry:New()
		return Person()
	end

	function ry:OnPush(obj)
		obj.Name = nil
	end

	per = ry()
	per.Name = "Ann"

	ry(per)

	per = ry()
	print(per.Name) -- Anonymous
end)
```

For simple types, the **Recycle** has provided a default **New** method, normally, we don't need to override the **New** method. We can provide the type and construction parameters when we create the recycle object:

```lua
require "PLoop" (function(_ENV)
	class "Person" (function(_ENV)
		function __ctor(self, name)
			print("New person", name)
			self.Name = name
		end
	end)

	ry = Recycle(Person, "Anonymous%d")

	per = ry() -- New person	Anonymous1
	per = ry() -- New person	Anonymous2
end)
```

If a parameter contains a '%d' pattern, it'd be converted as the index number, it's a simple feature provided by the **Recycle**, if it can't meet your require, you always can override the **New** method to support your types.


[Openresty]: https://github.com/openresty/lua-nginx-module/ "Openresty"
[NgxLua]: https://github.com/kurapica/NgxLua/ "An implementation for the Openresty"