# Exception

In Lua, there are two kinds exceptions: expected and un-expected. We use *error* api to raise expected error to a properly stack level, and we may face un-expected error like spell error `prnit(123)`.

The un-expected error is good to remain where it raised so we can simply solve it. But in a complex system like **PLoop**, the properly stack level of an expected error isn't always clear.

Take an example of class:

```lua
require "PLoop" (function(_ENV)
	class "A" {
		function (self)
			error("Something error", 5)
		end
	}

	o = A()
end)
```

If we need throw error in a constructor, we need use 5 as the properly stack level, it not good for the users, and if we create a new class inherit it:

```lua
require "PLoop" (function(_ENV)
	class "A" {
		function (self)
			error("Something error", 5)
		end
	}

	class "B" (function(_ENV)
		inherit "A"

		function B(self)
			super(self)
		end
	end)

	o = B()
end)
```

The stack level won't be right, since we can't pass the stack level through those constructors. And with the overload system, the stack level would be more confusion.

Since Lua can use table as error message if the function is called through *pcall*, the **PLoop** provide an simple exception system.


## System.Exception & throw

**System.Exception** objects represent errors that occur during application execution, it's the root class of the exception system, although the **PLoop** only provided two exception types, the other is **System.UnitTest.TestFailureException**, we'll learn it in the unit test system.

Property             |Description
:--------------------|:--------------------------------
Message              |a message that describes the current exception
StackTrace           |a string representation of the immediate frames on the call stack
TargetSite           |the method that throws the current exception
Source               |the source of the exception
InnerException       |the Exception instance that caused the current exception
Data                 |key/value pairs that provide additional information about the exception
LocalVariables       |key/value pairs of the local variable
Upvalues             |key/value pairs of the upvalues
StackDataSaved       |whether the stack data is saved, the system will save the stack data if the value is false when the exception is thrown out
StackLevel           |the stack level to be scanned, default 1, where the throw is called
SaveVariables        |whether save the local variables and the upvalues for the exception, default false

The constructor of the **Exception** is:

```lua
__Arguments__{ String, Exception/nil, Boolean/nil }
function Exception(self, message, inner, savevariables)
    self.Message        = message
    self.InnerException = inner
    self.SaveVariables  = savevariables
end
```

So we can create it only with an error message. The other values like the source, upvalues, localvariables are generated when the exception is thrown away:

```lua
require "PLoop" (function(_ENV)
	function test(a)
		local v = 2
		throw(Exception("Something error", nil, true))
	end

	local ok, exp = pcall(test, 1)

	-- Something error
	print(exp.StackTrace:match("^[^\n]*"))

	-- xxxx.lua:4
	print(exp.Source)

	-- a	1
	-- v	2
	for k, v in pairs(exp.LocalVariables) do
		print(k, v)
	end
end)
```

If we only want throw an error, don't need to track the variables, we also can pass string to the throw keyword, it'd generated an exception based on the string.

```lua
require "PLoop" (function(_ENV)
	class "A" (function(_ENV)
		function A(self)
			throw("Something error")
		end
	end)

	-- xxxx.lua:9: Something error
	o = A()
end)
```

The object creation is controlled by the system, so the **throw** mechanism is fully supported.

BTW, if you want all exception contains the local variables and up values, you can turn on the **EXCEPTION_SAVE_VARIABLES** in the **PLOOP_PLATFORM_SETTINGS**:

```lua
PLOOP_PLATFORM_SETTINGS = { EXCEPTION_SAVE_VARIABLES = true }
require "PLoop" (function(_ENV)
	function test(a)
		local v = 2
		throw("Something error")
	end

	local ok, exp = pcall(test, 1)

	-- Something error
	print(exp.StackTrace:match("^[^\n]*"))

	-- xxxx.lua:4
	print(exp.Source)

	-- a	1
	-- v	2
	for k, v in pairs(exp.LocalVariables) do
		print(k, v)
	end
end)
```


## Support throw

It's very simple to add **Exception** and **throw** to your codes:

```lua
require "PLoop" (function(_ENV)
	function task()
		throw("Something error")
	end

	function test()
		local ok, ret = pcall(task)

		if not ok then
			if type(ret) == "string" then
				error(ret, 0)           -- the un-expected error, remain where it raised
			else
				error(tostring(ret), 2) -- the expected error, raised to the caller
			end
		end
	end

	-- xxxx.lua:19: Something error
	test()
end)
```


## `__Arguments` & throw

The `__Arguments__` attribute will wrap the target function to a new one, since the argument check can't take all exceptions, we still need raise error in the method, we can decalre the attribute as throwable so the attribute will make sure the inner exception is handled.

```lua
require "PLoop" (function(_ENV)
	__Arguments__{ Number }:Throwable()
	function div(num)
		if num == 0 then
			throw("zero is not allowed")
		end
	end

	-- xxxx.lua:10: zero is not allowed
	div(0)
end)
```


## Property & throw

We also could throw exception in a property's get/set method, since there are plenty settings for the property, we can't make sure which is the properly stack level, so it's better to leave it to the property system:


```lua
require "PLoop" (function(_ENV)
	class "Person" (function(_ENV)
		field { __name = "anonymous" }

		function SetName(self, name)
			if type(name) ~= "string" then
				throw("The name must be string")
			end
			self.__name = name
		end

		function GetName(self)
			return self.__name
		end

		property "Name" {
			get = "GetName", -- or getmethod = "GetName"
			set = "SetName", -- or setmethod = "SetName"
			throwable = true,
		}
	end)

	-- xxxx.lua:24: The name must be string
	Person().Name = 123
end)
```