# AutoClose & with

When we need open some resources like file, data base, we must close them no matter how the operation processed.

Take a file as example:

```lua
function calcSum(path)
	local f = io.open(path, "r")
	local s = 0

	while l = f:lines() do
		s   = tonumber(l) + s
	end

	f:close()

	return s
end
```

If a line don't contains the number value, an error will be raised and the file won't be closed.


## System.IAutoClose

The **IAutoClose** is an interface only contains two abstract method that need to be override:

Abstract Method             |Arguments                |Description
:---------------------------|:------------------------|:-----------------------------
Open                        |                         |Process the open operation
Close                       |error                    |Process the close operation with error message(if failed)


## with

 The **with** keyword can receive several IAutoClose objects to auto open and close them no matter if there are errors:

```lua
require "PLoop" (function(_ENV)
	class "A" { IAutoClose,

		Open = function(self)
			print( "Open " .. self.name )
		end,

		Close = function(self, err)
			print("Close " .. self.name .. (err and (" with " .. err) or " without error"))
		end,
	}

	-- Open task
	-- process task
	-- Close task without error
	with(A{ name = "task"})(function(obj)
		print("process " .. obj.name)
	end)

	-- Open task
	-- Open task2
	-- process task, task2
	-- Close task with xxx.lua:23: 2333
	-- Close task2 with xxx.lua:23: 2333
	-- Catch error:xxx.lua:23: 2333
	with(A{ name = "task"}, A{ name = "task2"})(function(obj, obj2)
		print("process " .. obj.name .. ", " .. obj2.name)
		error("2333")
	end, function(err)
		print("Catch error:" .. err)
	end)
end)
```

So the second function would be used as error handler, if ommit, the *error* api will be used.

You can get all return values from the inner function:

```lua
require "PLoop" (function(_ENV)
	class "A" { IAutoClose }

	-- 1	2	3	4
	print(with(A())(function(obj)
		return 1, 2, 3, 4
	end))
end)

```

Here is a real project example for data base operation(see System.Data for more examples):

```lua
function RecordLastLogin(id)
	-- New database context and open the connection
	with(MyDBContext())(function(ctx)
		-- Process a transaction
		with(ctx.Transaction)(function(trans)
			-- Query and lock the data row
			local user = ctx.Users:Lock{ id = id }:First()
			if user then
				user.LastLogin = Date.Now

				-- save to the data base
				ctx:SaveChanges()
			else
				-- cancel the transaction
				trans:Rollback()
			end
		end)
	end)
end
```