# System.IO

The **System.IO** is a standalone lib, so to use it we must use `require "PLoop.System.IO"` to load it.

It provided several IO operations like get files from directory, read/write file and etc.

Since this is a pure Lua lib, all operations are created based on the `io.popen`, it's not designed for heavy IO scenarios and it's not the core part of the **PLoop**, so you always can replace it with the 3rd party libs(It's better to keep the type features in the IO lib, just replace its logic part, so other libs don't need to be changed).


## Table of Contents

* [System.IO.OperationSystem](#systemiooperationsystem)
* [System.IO.Path](#systemiopath)
	* [GetCurrentPath](#getcurrentpath)
	* [IsPathRooted](#ispathrooted)
	* [GetPathRoot](#getpathroot)
	* [GetDirectory](#getdirectory)
	* [CombinePath](#combinepath)
	* [GetSuffix](#getsuffix)
	* [GetFileName](#getfilename)
	* [GetFileNameWithoutSuffix](#getfilenamewithoutsuffix)
* [System.IO.Directory](#systemiodirectory)
	* [GetDirectories](#getdirectories)
	* [GetFiles](#getfiles)
	* [Exist](#exist)
	* [Create](#create)
	* [Delete](#delete)
* [System.IO.File](#systemiofile)
	* [GetCreationTime](#getcreationtime)
	* [GetLastWriteTime](#getlastwritetime)
	* [GetLastAccessTime](#getlastaccesstime)
	* [Exist](#exist)
	* [Delete](#delete)
* [System.IO.TextWriter](#systemiotextwriter)
* [System.IO.TextReader](#systemiotextreader)
* [System.IO.FileWriter](#systemiofilewriter)
* [System.IO.FileReader](#systemiofilereader)
* [System.IO.StringWriter](#systemiostringwriter)
* [System.IO.StringReader](#systemiostringreader)


## System.IO.OperationSystem

Before define other types, we need know which OS the Lua is running, the **OperationSystem** only provide a static property **OperationSystem.Current** to retrieve the os type, it'd be one of the **System.IO.OperationSystemType** :

* Unknown   - 0
* Windows   - 1
* MacOS     - 2
* Linux     - 4


## System.IO.Path

The **Path** is a tool class used to get current path, combine path, get file name from path and etc.

### GetCurrentPath

Get the execution path of the Lua:

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetCurrentPath())
```

### IsPathRooted

Check if the path contains the root path

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.IsPathRooted("/www/html")) -- true
```


### GetPathRoot

Gets the root of the path

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetPathRoot("C:\\www\\html")) -- C:\
```


### GetDirectory

Gets the directory of the path, if the path is a root, empty string would be returned

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetDirectory("C:\\www\\html")) -- C:\www
```


### CombinePath

Combine the paths

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.CombinePath("C:\\www\\html", "controller", "home.lua")) -- C:\www\html\controller\home.lua
```


### GetSuffix

Gets the suffix of the path include the "."

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetSuffix("C:\\www\\html\\img\\test.jpg")) -- .jpg
```


### GetFileName

Gets the file name with the suffix

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetFileName("C:\\www\\html\\img\\test.jpg")) -- test.jpg
```


### GetFileNameWithoutSuffix

Gets the file name without the suffix

```lua
require "PLoop"
require "PLoop.System.IO"

print(PLoop.System.IO.Path.GetFileNameWithoutSuffix("C:\\www\\html\\img\\test.jpg")) -- test
```


## System.IO.Directory

The **Directory** is a tool class used to provde get sub directories or files, create, delete directory operations.

### GetDirectories

Gets the directories of the target directory

```lua
require "PLoop"
require "PLoop.System.IO"

import("System.IO", true)  -- import IO and what in it to the _G

for name in Directory.GetDirectories(Path.GetCurrentPath()) do
	print(name)
end
```


### GetFiles

Gets the files in the target directory

```lua
require "PLoop"
require "PLoop.System.IO"

import("System.IO", true)

for name in Directory.GetFiles(Path.GetCurrentPath()) do
	print(name)
end
```


### Exist

Check if the target directory is existed

```lua
require "PLoop"
require "PLoop.System.IO"

import("System.IO", true)

```


### Create

Create a directory


### Delete

Delete the directory

```lua
require "PLoop"
require "PLoop.System.IO"

import("System.IO", true)

local path = Path.CombinePath(Path.GetCurrentPath(), "test")

print(Directory.Exist(path)) -- false
Directory.Create(path)
print(Directory.Exist(path)) -- true
Directory.Delete(path)
print(Directory.Exist(path)) -- false
```


### System.IO.File

The **File** is a tool class used to provide file operations:


### GetCreationTime

Try to get the file's creation time (in string format)


### GetLastWriteTime

Try to get the file's last modified time (in string format)


### GetLastAccessTime

Try to get the file's last access time (in string format)


### Exist

Check if the file existed


### Delete

Delete the file


```lua
require "PLoop"
require "PLoop.System.IO"

import("System.IO", true)

local path = Path.CombinePath(Path.GetCurrentPath(), "testio.lua")

print(File.Exist(path)) -- false

local f = io.open(path, "w")
f:write("hi")
f:close()

print(File.GetCreationTime(path))

print(File.Exist(path)) -- true
File.Delete(path)
print(File.Exist(path)) -- false
```


## System.IO.TextWriter

The **TextWriter** represents writers that can write a sequential series of characters. It's an abstract class defined the abstract features for all writer classes.

It's also extend the **System.IAutoClose** interface, so the writer class should also define the **Open** and **Close** method, and can be used under **with** keyword.

Abstract Property      |Type                    |Description
:----------------------|:-----------------------|:---------------------
Encoding               |System.Text.Encoding    |Gets the character encoding in which the output is written
NewLine                |String                  |Gets or sets the line terminator string used by the current TextWriter. Default "\n"


Abstract Method        |Arguments               |
:----------------------|:-----------------------|:---------------------
Flush                  |                        |Clears all buffers for the current writer and causes any buffered data to be written to the underlying device.
Write                  |data:String             |Writes the data to the text string or stream.


Method                 |Arguments               |
:----------------------|:-----------------------|:---------------------
WriteLine              |data:String             |Writes the data(could be nil) followed by a line terminator to the text string or stream.


There are two writer class defined: **System.IO.FileWriter**, **System.IO.StringWriter**.


## System.IO.TextReader

The **TextReader** represents readers that can read a sequential series of characters. It's an abstract class defined the abstract features for all reader classes.

It's also extend the **System.IAutoClose** interface, so the reader class should also define the **Open** and **Close** method, and can be used under **with** keyword.


Abstract Property      |Type                    |Description
:----------------------|:-----------------------|:---------------------
Encoding               |System.Text.Encoding    |Gets the character encoding in which the output is written
Position               |Number                  |Gets or Sets the operation position


Abstract Method        |Arguments               |
:----------------------|:-----------------------|:---------------------
Read                   |                        |Reads the next character from the text reader and advances the character position by one character.
ReadLine               |                        |Reads a line of characters from the text reader and returns the data as a string.
ReadBlock              |                        |Reads a specified maximum number of characters from the current text reader and writes the data to a buffer, beginning at the specified index.
ReadToEnd              |index, count            |Reads all characters from the current position to the end of the text reader and returns them as one string.


Method                 |Arguments               |
:----------------------|:-----------------------|:---------------------
ReadLines              |                        |Return the ReadLine method and self for a generic for


There are two writer class defined: **System.IO.FileReader**, **System.IO.StringReader**.


## System.IO.FileWriter

The **FileWriter** is a class based on the **io.open**, it inherit the **TextWriter** and provide the ability to write text to files.

The constructor is

```lua
__Arguments__{ String, FileWriteMode/FileWriteMode.Write }
function FileWriter(self, file, mode)
end
```

The **System.IO.FileWriter.FileWriteMode** is an enum contains the mode of file writing:

* Write               - mode "w"
* Append              - mode "a"
* WritePlus           - mode "w+"
* AppendPlus          - mode "a+"
* WriteBinary         - mode "wb"
* AppendBinary        - mode "ab"
* WritePlusBinary     - mode "w+b"
* AppendPlusBinary    - mode "a+b"


## System.IO.FileReader

The **FileReader** is a class based on the **io.open**, it inherit the **TextReader** and provide the ability to read text from files.

The constructor is

```lua
__Arguments__{ String, FileReadMode/FileReadMode.Read }
function FileWriter(self, file, mode)
end
```

The **System.IO.FileWriter.FileWriteMode** is an enum contains the mode of file writing:

* Read                - mode "r"
* ReadBinary          - mode "rb"


## System.IO.StringWriter

The **StringWriter** is a class based on the **io.open**, it inherit the **TextWriter** and provide the ability to write text to files.

Here is an example:

```lua
require "PLoop"
require "PLoop.System.IO"

PLoop(function(_ENV)
	import "System.IO"

	local sw = StringWriter()

	with(sw)(function(sw)
		sw:Write("Test")
		sw:Write(" ")
		sw:Write("is done")
	end)

	print(sw.Result) -- Test is done
end)
```


## System.IO.StringReader

The **StringReader** is a class based on the **io.open**, it inherit the **TextReader** and provide the ability to write text to files.

Here is an example:

```lua
require "PLoop"
require "PLoop.System.IO"

PLoop(function(_ENV)
	import "System.IO"

	local sw = StringReader("This is \na test")

	-- This is
	-- a test
	for line in sw:ReadLines() do
		print(line)
	end
end)
```

Those reader & writer are normally used within the serialization system.