# System.IO.Resource

Besides the normal Lua files, we still need others, like a html template file to generate http response. They are resources, and the **Resource** system is used to provide support for it.

**This is a topic only usable for framework creators, so you can just skip it if you don't need to know the details behind the systems like the web framework.**


## System.IO.Resource.IResourceLoader

Every type resources should have a resource loader to load them into the system, like build a view class from html template.

The **IResourceLoader** is the root interface for all resource loaders, also the initiator that start loading resources.


Static Method          |Arguments                                                               |Description
:----------------------|:-----------------------------------------------------------------------|:-----------------------
LoadResource           |path:String, reader:TextReader/nil, env:Table/nil, tryLock:Boolean/nil  |Load the resource of given path


Abstract Method        |Arguments                                                               |Description
:----------------------|:-----------------------------------------------------------------------|:-----------------------
Load                   |path:String, reader:TextReader/nil, env:Table/nil                       |Load the resource from the path


To declare a resouece loader class, we could use the `System.IO.Resource.__ResourceLoader__` attribute, here is an example used by **System.IO.Resource.LuaLoader** which used to load the lua file and return the type with the file's name:

```lua
System.IO.Resource.__ResourceLoader__"lua"
class "System.IO.Resource.LuaLoader" { System.IO.Resource.IResourceLoader }
```

So we only need create class extend the **IResourceLoader** and bind with suffix, if you need the loader match several suffix, you can use `__ResourceLoader__("lua", "luax", "luac")`.

Here is an example to load a resource:

```lua
require "PLoop"
require "PLoop.System.IO"

PLoop(function(_ENV)
	import "System.IO"
	import "System.IO.Resource"

	code = [[
		PLoop(function(_ENV)
			class "Test" {}
		end)
	]]

	-- The path is needed so the system can get the suffix and diff it from others
	cls = IResourceLoader.LoadResource("test.lua", StringReader(code))

	print(Class.Validate(cls)) -- Test
end)
```

So we load a code named `test.lua`, and the class *Test* defined in it should be returned. You may read the **LuaLoader** for more details.


## System.IO.Resource.IResourceManager

Besides loading resources, we also need to manager them, we should keep using the loaded resources, and may reload them when their file is modified.

The **IResourceManager** provided the resource controls based on the **IResourceLoader**.

Static Method          |Arguments                      |Description
:----------------------|:------------------------------|:----------------------
LoadResource           |path:String, env:Table/nil     |Load the file and return the resource
GetResourcePath        |resource:Any                   |Gets the file path from a resource
AddRelatedPath         |path:String, relative:String   |Add a related path to the resource path for reload checking
SetReloadWhenModified  |path:String, flag:Boolean/true |Mark the path reload when modified


Take the web template system as an example, a page class may use another page as it's master, so it don't need to set the sharable parts like the navigation bars, the master page is a related path to it, so when the master page is modified, when load the page class again, the related paths should also be checked, the master page'll be reloaded and the page class will get the update.


Static Property        |Description
:----------------------|:----------------------
ReloadWhenModified     |whether reload all the file when modified, default false

If we turn on the property, all registered resources should be reloaded when modified, it's useful in development, should be turn off when released.