# System.Data

The **System.Data** provide a data entity framework. It's an abstract layer between the data entity objects and the database. To use it on any data sources, you should have or create the implementation of several interfaces. We'll see more details at later.

## Table of Contents

* [Data Context and Data Entity](#data-context-and-data-entity)
	* [System.Data.IDataContext](#systemdataidatacontext)
	* [System.Data.IDataEntity](#systemdataidataentity)
	* [Data Fields](#data-fields)
	* [Query](#query)
	* [Create](#create)
	* [Update](#update)
	* [Delete](#delete)
	* [Type Converter](#type-converter)
	* [DBNull](#dbnull)
	* [Data View](#data-view)
* [Data Base Layer](#data-base-layer)
* [System.Data.ICache](#systemdataicache)
* [System.Data.DataEntityCache](#systemdatadataentitycache)


## Data Context and Data Entity

Above the abstract layer, we have the data context and data entity system.

The data entities are objects represents the datas from the data sources, normally they each stored a row data from a data table, it provides the ability to operate those datas through the code types, not the databse types.

The data context represents the source of all entities mapped over a database connection, it would track the changes of the entities generated from it, and the users can easily commit their changes by call the **SaveChanges** method. The data context **won't** cache the queried entities, we'll see how to use it with the cache system at later.

Task an example for a database to manage employees, first is the definition of the data context:

```lua
require "PLoop"
require "PLoop.System.Data"

PLoop(function(_ENV)
	import "System.Data"

	__DataContext__()
	class "TestDBContext" (function(_ENV)
		-- Init the connection when a context object is created
		function __ctor(self)
			self.Connection = MySQLConnection {
				host = "127.0.0.1",
				port = 3306,
				database = "TestDB",
				user = "root",
				password = "xxxxxx",
			}
		end

		-- Declare the entity classes for data tables
		__DataTable__{
			name         = "department",

			indexes      = {
				{ fields = { "id" },   primary = true },
			}
		}
		class "Department" (function(_ENV)
			-- autoincr means the id is auto given by the database when inserted
			__DataField__{ autoincr = true }
			property "id"           { type = NaturalNumber }

			__DataField__{ notnull = true, unique = true }
			property "name"         { type = String }
		end)

		__DataTable__ {
			name         = "employee",

			indexes      = {
				{ fields = { "id" },   primary = true },
				{ fields = { "name" },  unique = true },
				{ fields = { "telno" }, unique = true },
				{ fields = { "email" }, unique = true }
			}
		}
		class "Employee" (function(_ENV)
			-- autoincr means the id is auto given by the database when inserted
			__DataField__{ autoincr = true }
			property "id"           { type = NaturalNumber }

			__DataField__{ notnull = true }
			property "name"         { type = String }

			__DataField__{ notnull = true }
			property "joindate"     { type = Date }

			__DataField__{ }
			property "leavedate"    { type = Date }

			__DataField__()
			property "telno"        { type = String }

			__DataField__{ unique = true }
			property "email"        { type = String }

			-- The foreign field setting
			__DataField__{ foreign = { map = { depid = "id" }, link = { name = "Employees", order = "id" } } }
			property "department"   { type = Department }
		end)
	end)

	-- Write the below code here for test
end)
```

You can find the implementation to **MySQLProvider** in [NgxLua][] tested with mysql 8.0. We'll see a debug version at later.


### System.Data.IDataContext

When a class is marked with `__DataContext__` attribute, the class'll extend the **IDataContext** interface automatically, the interface has provided the basic features for data context objects.


Property                 |Type                        |Description
:------------------------|:---------------------------|:--------------------------
Connection               |System.Data.IDbConnection   |Sets or gets the Connection object to associate with the transaction, must provided when the data context is created
Transaction              |System.Data.IDbTransaction  |Gets an existed or new created transaction from the database connection for database operations


Method                   |Arguments                                                     |Description
:------------------------|:-------------------------------------------------------------|:------------------------------
Open                     |                                                              |Open the data connection
Close                    |                                                              |Close the data connection
SaveChanges              |                                                              |Save the data changes in the context
Query                    |sql: String, ... : Any * 0                                    |Query the format sql with parameters and return the result as a List
Execute                  |sql: String, ... : Any * 0                                    |Execute the format sql with parameters and return the result
QueryAsView              |viewcls:-System.Data.IDataView, sql: String, ... : Any * 0    |Send the query sql and return the result wrapped with the view
QueryView                |viewcls:-System.Data.IDataView, ... : Any * 0                 |Send the view's query sql and return the result wrapped with the view


The **Query**, **Execute** methods are used to run pure sql(or a format string to be sql), since a complex sql like JOIN operations are only used to do the query jobs, the system won't bring in any complex style like `ctx:From("user"):Join("email"):select(xxx)`, the sql is more simple and clear.

```lua
local ctx = TestDBContext()

ctx:Open()

local result = ctx:Query("SELECT * FROM department where id = %d", 100)

result:Each("d=>print(d.name)")

ctx:Close()
```

The **Query** result will be tables that saved in a **List** object, each table contains datas from a data row. The **Execute** methods will return anything the database returned.

The disadvantage of this design is all data in those tables are simple type datas, like a date value will be in string format, can't be used directly.

This could be done by the **QueryAsView** and **QueryView** methods, but will be explained after the data context part.

When create a data context class, we need assign a data connection to the **Connection** property, normally this is done in the constructor.

Then we only need to declare the data table classes. The data view classes are only used to query, so they don't need to be part of the data context.


### System.Data.IDataEntity

The data table classes represents the data tables in database. They are simple classes that controlled by the data entity system, they only extend the **IDataEntity**(done by the `__DataTable__` attribute), their objects are the data entities, they are **serializable** by default.

The entities generated from the data table classes represents the data rows, each object contains the data from each data row.

The `__DataTable__` attribute only cares the data table settings:

```lua
-- The setting for table indexes
struct "DataTableIndex" {
	-- the index name, the system don't really check it, but it's useful if you want
	-- create the data tables based on the data context
    { name = "name",        type = String },

    -- whether the index is unqiue
    { name = "unique",      type = Boolean },

    -- whether the index is a fulltext index
    { name = "fulltext",    type = Boolean },

    -- whether the index is a primary index
    { name = "primary",     type = Boolean },

    -- the field list, just keep in mind, the names in it should be the fields of
    -- the data table, not the table class
    { name = "fields",      type = struct { String } },
}

-- The setting for the data table
struct "DataTableSetting" {
	-- the data table's name, could use the class's name if omitted
    { name = "name",        type = String },

    -- the data table's indexes
    { name = "indexes",     type = struct { DataTableIndex } },

    -- After defined in the data context, the data context will automatically
    -- create a readonly property of **DataCollection[tableCls]** for easy using,
    -- normally the property name should be the class's name + "s", for *Employee*,
    -- that should be *Employees*, if you mind, you can give a name here to replace it.
    { name = "collection",  type = String },

    -- the database engine of the data table, normally useless to the system, but useful when creating the databases.
    { name = "engine",      type = String },
}
```

The attribute use the **DataTableSetting** as its parameter, since they are all optional, you can also pass no paramter(which the *name* is the data table class's name) or a string (used as the *name*).

With the data table settings, the system will know the primary index of the data table, and make features based on that.(The **DataCollection** template will make improvement based on the target class's settings)

The **IDataEntity** interface has provided several methods(useful for outside):

Method                  |Description
:-----------------------|:-------------------
Delete                  |Mark self as deleted, the data row will be deleted when the context save changes


### Data Fields

The data table keys are mapped to properties of the entity classes. With the property, the system would know when an entity has changed.

So we use the `__DataField__` attribute to declare the property is a map from the data key, the attribute also require a complex **System.Data.FieldSetting**.

```lua
-- The query order like { name = "id", desc = true } -> "ORDER BY id DESC"
struct "QueryOrder" {
	{ name = "name",   type = String, require = true },
	{ name = "desc",   type = Boolean },
}

-- The query order list { { name = "id", desc = false }, { name = "age", desc = true} } -> "ORDER BY id, age DESC"
struct "QueryOrders" { String + QueryOrder }

-- Convert the value between the code type and data type, normally no need
-- to use it, the default converter is provided by the framework
struct "TypeConverter" {
	-- data type -> code type
	{ name = "fromvalue", type = Function, require = true },

	-- code type -> date type
	{ name = "tovalue",   type = Function, require = true },

	-- the second value passed to the fromvalue and tovalue function
	{ name = "format",    type = Any }
}

-- The primary link settings
struct "PrimaryLink" {
	{ name = "name",    type = String, require = true },
	{ name = "order",   type = String + QueryOrders },
}

-- The foreign map settings
struct "ForeignMap" {
	-- The fields map between the foreign key to the primary key in other table
	{ name = "map",     type = Table, require = true },

	-- The primary link settings
	{ name = "link",    type = String + PrimaryLink },
}

-- The field settings
struct "FieldSetting" {
	-- The data table key name, use the property name is omit
	{ name = "name",        type = String },

	-- The data type of the key, used when create or update the data table schema
	{ name = "type",        type = String },

	-- Whether it's a unique key
	{ name = "unique",      type = Boolean },

	-- Whether it's an auto increase key, it's value is generated by database when inserted
	{ name = "autoincr",    type = Boolean },

	-- Whether the key's data can't be nil
	{ name = "notnull",     type = Boolean },

	-- The foreign settings
	{ name = "foreign",     type = ForeignMap },

	-- The code type and data type's value converter, use the default converter is omit
	{ name = "converter",   type = TypeConverter },

	-- The special format for converter
	{ name = "format",      type = Any },
}
```

For normal data table keys, the settings only need to set the *name*, *type*, *unique*, *autoincr* and *notnull*, it's very simple, you can find them in the previous examples, the hard part is the foreign key settings.

In the *Employee* entity class, there is a foreign key settings like :

```lua
-- The foreign field setting
__DataField__{ foreign = { map = { depid = "id" }, link = { name = "Employees", order = "id" } } }
property "department"   { type = Department }
```

* The property type is the Department, that means the primary key is in the *Department* entity class

* `map = { depid = "id" }`, means the foreign key in the data table is the *depid*, it's mapped to the primary key *id* of the **department** data table(all keys in the data field settings are keys in data table, not the property name in the entity class).

* `link = { name = "Employees", order = "id" }` means add a **Employees** property in the **Department** entity class, so any department entity can use its **Employees** property to retrieve all employees entities of the department, the *id* key will be used as the order. If the foreign key is unique, the **department.Employees** will return only one entity, otherwise, it will return a **List** of the **Employee** entities. Since a department can't have only one employee, it'd be not unique.


### Query

Now we have the data context **TestDBContext**, it has two entity classes, here is the query usage of them:

```lua
local ctx = TestDBContext()

-- Open the database connection
ctx:Open()

-- Query a department whose id is 1
local depart = ctx.Departments:Query{ id = 1 }:First()

-- Gets all its employees
for _, employee in depart.Employees:GetIterator() do
	print(employee.name)
end

-- Query a employee
local employee = ctx.Employees:Query{ name = "Ann" }:First()

-- Get its deparment name
print(employee.department.name)

-- Close the database connection
ctx:Close()
```

The `ctx.Departments` is an object of **DataCollection[TestDBContext.Department]**, the **DataCollection** is a template class, would adjust its codes based on the entity classes, it provide several methods:


Method              |Arguments                                |Description
:-------------------|:----------------------------------------|:----------------------------------
Query               |condition: Table, order:QueryOrders/nil  |Query the data with orders(if existed) based on the condition, save the results as the entity class's objects
Query               |sql : String, ... : Any * 0              |Query the data with sql and save the results as the entity class's objects
Lock                |condition: Table, order:QueryOrders/nil  |Lock and query the data with orders(if existed) based on the condition, save the results as the entity class's objects, the lock operation may failed, an error would be raised
QueryAll            |order:QueryOrders/nil                    |Query all data with orders(if existed), save the results as the entity class's objects
Add                 |data: Table                              |Warp and return the data as the entity class's object, it'll be inserted when the context save changes

Unlike the **Query** method of the **IDataContext**, the data collection are focused on data entities, it take a data row as a single entity. So normally this is used to do the create, update, delete jobs.

The **Query** always return a list (even no result), so we always can use the **First** method if we only expect one result.

The `depart.Employees` and `employee.department` are properties, they'll query the database and generate the entities for the first time accessing.

Now we can have more examples for CRUD:


### Create

It's not safe to manully call the **Open** and **Close** method, since there may be error during the process, we can use the **with** keyword to make sure they'll be safely closed.

```lua
-- Create data row
function addEmployee(deptid, name, telno, email)
	local employee

	with(TestDBContext())(function(ctx)
		-- Open the database connection
		local dept = ctx:Query{ id = deptid }:First()

		if not dept then return end

		with(ctx.Transaction)(function(trans)
			-- Start a transaction

			-- Add a new employee to the data table
			employee    = ctx.Employees:Add{
				name 	= name,
				telno 	= telno,
				email 	= email,
				joindate= Date.Now,
			}

			-- The foreign property should be assigned outside the Add method
			employee.department = dept

			-- Save the data to database
			ctx:SaveChanges()
		end)
	end)

	-- we can get employee id after it's inserted
	-- it also could be nil, if insert failed
	return employee and employee.id
end
```


### Update

We must lock the data row before update it:

```lua
--- Update a data row
function fireEmployee(id)
	if not id then return end

	with(TestDBContext())(function(ctx)
		-- Open the database connection

		with(ctx.Transaction)(function(trans)
			-- Start a transaction

			-- Query and Lock the employee
			local employee = ctx.employees:Lock{ id = id }:First()

			if employee the
				employee.leavedate = Date.Now

				-- Save the operation to database
				ctx:SaveChanges()
			else
				-- Cancel the transaction
				trans:Rollback()
			end
		end)
	end)
end
```


### Delete

Also you can just use the **Execute** method of the data context.

```lua
-- Delete a data row
function deleteEmployee(id)
	if not id then return end

	with(TestDBContext())(function(ctx)
		-- Open the database connection

		with(ctx.Transaction)(function(trans)
			-- Start a transaction

			-- Query and Lock the employee
			local employee = ctx.employees:Lock{ id = id }:First()

			if employee the
				employee:Delete()

				-- Save the operation to database
				ctx:SaveChanges()
			else
				-- Cancel the transaction
				trans:Rollback()
			end
		end)
	end)
end
```

When an entity is modified, it'll be changed to un-modified when the data context has saved changes, so it won't be saved in the next **SaveChanges** call.


### Type Converter

For common value types, string and number should be easily mapped, but it's a little different for types like **Boolean**, **Date** and etc. The **Date** value should be string that get/set to the database, it's easy to transfer, but hard to use, the **Boolean** should be saved as 1(true) and 0(false) as TINYINT(for mysql), also hard to be used in Lua(0 is also true).

The data entity use property to manipulate the data values, when we read it, it can convert the data type value to our code type value, and when we write it, the code type value should be changed to the data type value.

We already see it in the **DataFieldSetting**, take the **Date** as example:

```lua
local ctx = TestDBContext()

-- Open the database connection
ctx:Open()

-- Query a employee
local ann = ctx.Employees:Query{ name = "Ann" }:First()

-- print the day of year
print(ann.joindate.DayOfYear)

ann.joindate = ann.joindate:AddDays(100)

-- Close the database connection
ctx:Close()
```

So we can use the *joindate* as code type **Date** directly. This is done by a static setting:

```lua
__DataField__.Converter[Date] = {
    fromvalue       = Date.Parse,
    tovalue         = Date.ToString,
}
```

The format is not provided, since the date format in **Mysql** is the same to the **Date**'s default format `"%Y-%m-%d %X"`.

If the data type in database require a special format like `"%Y-%m-%d"`, we should change the `__DataField__` settings like :

```lua
__DataField__{ notnull = true, format = "%Y-%m-%d" }
property "joindate"     { type = Date }
```

The **Boolean** also have a converter that used to map 1 to true, 0 to false. You may use the `__DataField__.Converter` to add your owns.


### DBNull

There also should have some special values represents the dbnull in the results, that must be converted to nil, otherwise codes like `if employee.leavedate then` will always true.

We should use the **System.Data.AddNullValue** to add those special values as dbnull:

```lua
System.Data.AddNullValue(ngx.null)  -- for OpenResty
```

You also can use the **System.Data.ParseValue** to filter the DBNull(return nil if it represents the DBNull).

```lua
value = System.Data.ParseValue(value)
```

We also need a constant value represents the db null, so we could know a data value is changed to null, not they don't existed. The **Data** also provide it - **System.Data.DBNull**, with the data entity system, you don't really need to use it.


### Data View

The results from the **IDataContext.Query** will keep their data types(also with their special DBNull value), if the query schema is fixed, we can provide a data view for the query results, so we can access them like the a data entity.

```lua
__DataView__[[select e.id, e.name, d.name as deptname FROM employee e, department d where e.id = %d and e.depid = d.id]]
class "EmployeeView" (function(_ENV)
		__DataField__()
		property "id"           { type = NaturalNumber }

		__DataField__()
		property "name"         { type = String }

		__DataField__()
		property "deptname"  	{ type = String }
end)
```

The data view can use a sql as its view sql, so we can query it only with the parameters:

```lua
local e = ctx:QueryView(EmployeeView, 100):First()
print(e.id, e.name, e.deptname)
```

Also we can use **QueryAsView** for a specific sql.


## Data Base Layer

Below the abstract layer, we need defined classes based on the database that implement three interfaces:

```lua
enum "TransactionIsolation" {
	REPEATABLE_READ         = 0,
	READ_UNCOMMITTED        = 1,
	READ_COMMITTED          = 2,
	SERIALIZABLE            = 3,
}

interface "ISqlBuilder" (function(_ENV)
	-----------------------------------------------------------
	--                        method                         --
	-----------------------------------------------------------
	--- Sets the select fields
	-- @param fields    a list of the field or the string that contains the fields
	-- @return self
	__Abstract__() function Select(self, fields) return self end

	--- Lock the query rows
	-- @return self
	__Abstract__() function Lock(self) return self end

	--- Sets the updating field-value map
	-- @param map       a map for field to value
	-- @return self
	__Abstract__() function Update(self, map) return self end

	--- Sets to delete data
	-- @return self
	__Abstract__() function Delete(self) return self end

	--- Insert the data
	-- @return self
	__Abstract__() function Insert(self, map) return self end

	--- Set the data table name
	-- @param name      the data table name
	-- @param alias     the alias name
	-- @return self
	__Abstract__() function From(self, name, alias) return self end

	--- Set the conditions
	-- @param condition the query condition
	-- @return self
	__Abstract__() function Where(self, condition) return self end

	--- Set the order by
	-- @param name      the data field name
	-- @param desc      whether use desc order
	-- @return self
	__Abstract__() function OrderBy(self, field, desc) return self end

	--- Generate the final sql
	-- @return sql
	__Abstract__() function ToSql(self) end
end)

--- Represents the connection to a data base
interface "IDbConnection" (function(_ENV)
	extend "IAutoClose"

	-----------------------------------------------------------
	--                       property                        --
	-----------------------------------------------------------
	--- The connection state
	__Abstract__() property "State"      { type = ConnectionState, default = ConnectionState.Closed }

	--- The query builder class
	__Abstract__() property "SqlBuilder" { type = -ISqlBuilder }

	-----------------------------------------------------------
	--                        method                         --
	-----------------------------------------------------------
	--- Get a new database transaction.
	__Abstract__() function NewTransaction(self, isolation) end

	--- Sends the query sql and return the result
	__Abstract__() function Query(self, sql, ...) end

	--- Sends the insert sql to the database and return the auto-increased id
	__Abstract__() function Insert(self, sql, ...) end

	--- Sends the update sql to the database
	__Abstract__() function Update(self, sql, ...) end

	--- Sends the delete sql to the database
	__Abstract__() function Delete(self, sql, ...) end

	--- Execute the insert sql and return the result
	__Abstract__() function Execute(self, sql, ...) end
end)

--- Represents a transaction to be performed at a data source
interface "IDbTransaction" (function(_ENV)
	extend "IAutoClose"  -- The Open & Close method is already implemented by IDbTransaction

	-----------------------------------------------------------
	--                       property                        --
	-----------------------------------------------------------
	--- The Connection object to associate with the transaction
	__Abstract__() property "Connection"        { type = IDbConnection }

	--- The transaction isolation level
	__Abstract__() property "Isolation"         { type = TransactionIsolation, default = TransactionIsolation.REPEATABLE_READ }

	-----------------------------------------------------------
	--                        method                         --
	-----------------------------------------------------------
	--- Begin the transaction
	__Abstract__() function Begin(self) end

	--- Commits the database transaction
	__Abstract__() function Commit(self) end

	--- Rolls back a transaction from a pending state
	__Abstract__() function Rollback(self) end
end)
```

Here is a simple example:

```lua
require "PLoop"
require "PLoop.System.Data"

PLoop(function(_ENV)
	import "System.Data"

	class "MySQLConnection" {}  -- define it later

	__Sealed__() class "MySQLTransaction" (function(_ENV)
		extend "IDbTransaction"

		-----------------------------------------------------------
		--                        method                         --
		-----------------------------------------------------------
		function Begin(self)
			self.Connection:Execute("TRANSACTION BEGIN")
		end
		--- Commits the database transaction
		function Commit(self)
			self.Connection:Execute("TRANSACTION COMMIT")
		end

		--- Rolls back a transaction from a pending state
		function Rollback(self)
			self.Connection:Execute("TRANSACTION ROLLBACK")
		end

		-----------------------------------------------------------
		--                      constructor                      --
		-----------------------------------------------------------
		__Arguments__{ MySQLConnection, TransactionIsolation/TransactionIsolation.REPEATABLE_READ }
		function __ctor(self, conn, isolation)
			self.Connection = conn
			self.Isolation  = isolation
		end
	end)

	__Sealed__() class "MySQLBuilder" (function(_ENV)
		extend "ISqlBuilder"

		export {
			SQLTYPE_SELECT      = 1,
			SQLTYPE_UPDATE      = 2,
			SQLTYPE_DELETE      = 3,
			SQLTYPE_INSERT      = 4,

			FIELD_SQLTYPE       = 1,
			FIELD_SELECT        = 2,
			FIELD_UPDATE        = 3,
			FIELD_INSERT        = 4,
			FIELD_FROM          = 5,
			FIELD_WHERE         = 6,
			FIELD_ORDERBY       = 7,

			quote_sql_str       = function(str) return ("%q"):format(str) end,
			type                = type,
			tblconcat           = table.concat,
			tostring            = tostring,
			pairs               = pairs,

			DBNull
		}

		local function escape(val)
			if val == DBNull then return "NULL" end

			local vtype         = type(val)

			if vtype == "boolean" then
				return val and "1" or "0"
			elseif vtype == "string" then
				return quote_sql_str(val)
			else
				return tostring(val)
			end
		end

		-----------------------------------------------------------
		--                        method                         --
		-----------------------------------------------------------
		function Select(self, fields)
			self[FIELD_SQLTYPE] = SQLTYPE_SELECT

			if type(fields) == "table" then
				fields          = tblconcat(fields, ",")
			else
				fields          = tostring(fields)
			end

			self[FIELD_SELECT]  = fields ~= "" and fields or nil

			return self
		end

		function Insert(self, map)
			if type(map) == "table" then
				self[FIELD_SQLTYPE] = SQLTYPE_INSERT

				local fields    = {}
				local values    = {}
				local index     = 1

				for fld, val in pairs(map) do
					fields[index] = fld
					values[index] = escape(val)
					index       = index + 1
				end

				fields          = tblconcat(fields, ",")
				values          = tblconcat(values, ",")

				if fields ~= "" then
					self[FIELD_SELECT] = fields
					self[FIELD_INSERT] = values
				end
			end

			return self
		end

		function Update(self, map)
			self[FIELD_SQLTYPE] = SQLTYPE_UPDATE

			if type(map) == "table" then
				local temp      = {}
				local index     = 1
				local first     = true

				for fld, val in pairs(map) do
					if not first then
						temp[index] = ","
						index   = index + 1
					end
					first       = false

					temp[index] = fld
					index       = index + 1

					temp[index] = "="
					index       = index + 1

					temp[index] = escape(val)
					index       = index + 1
				end

				map             = tblconcat(temp, "")
			else
				map             = tostring(map)
			end

			self[FIELD_UPDATE]  = map ~= "" and map or nil

			return self
		end

		function Delete(self)
			self[FIELD_SQLTYPE] = SQLTYPE_DELETE

			return self
		end

	   function From(self, name)
			self[FIELD_FROM]    = name
			return self
		end

		function Where(self, condition)
			if type(condition) == "table" then
				local temp      = {}
				local index     = 1
				local first     = true

				for fld, val in pairs(condition) do
					if not first then
						temp[index] = " AND "
						index   = index + 1
					end
					first       = false

					temp[index] = fld
					index       = index + 1

					temp[index] = "="
					index       = index + 1

					temp[index] = escape(val)
					index       = index + 1
				end

				condition       = tblconcat(temp, "")
			else
				condition       = tostring(condition)
			end

			self[FIELD_WHERE]   = condition ~= "" and condition or nil

			return self
		end

		function OrderBy(self, field, desc)
			if desc then field  = field .. " DESC" end

			if self[FIELD_ORDERBY] then
				self[FIELD_ORDERBY] = self[FIELD_ORDERBY] .. "," .. field
			else
				self[FIELD_ORDERBY] = field
			end

			return self
		end

		function ToSql(self)
			local temp          = {}

			local sqltype       = self[FIELD_SQLTYPE]

			if not sqltype then return end

			if sqltype == SQLTYPE_SELECT then
				temp[1]         = "SELECT"
				temp[2]         = self[FIELD_SELECT] or "*"
				temp[3]         = "FROM"

				if not self[FIELD_FROM] then return end
				temp[4]         = self[FIELD_FROM]

				if self[FIELD_WHERE] then
					temp[5]     = "WHERE"
					temp[6]     = self[FIELD_WHERE]
				else
					temp[5]     = ""
					temp[6]     = ""
				end

				if self[FIELD_ORDERBY] then
					temp[7]     = "ORDER BY"
					temp[8]     = self[FIELD_ORDERBY]
				else
					temp[7]     = ""
					temp[8]     = ""
				end
			elseif sqltype == SQLTYPE_UPDATE then
				temp[1]         = "UPDATE"

				if not self[FIELD_FROM] then return end
				temp[2]         = self[FIELD_FROM]
				temp[3]         = "SET"

				if not self[FIELD_UPDATE] then return end
				temp[4]         = self[FIELD_UPDATE]

				if self[FIELD_WHERE] then
					temp[5]     = "WHERE"
					temp[6]     = self[FIELD_WHERE]
				else
					temp[5]     = ""
					temp[6]     = ""
				end
			elseif sqltype == SQLTYPE_DELETE then
				temp[1]         = "DELETE FROM"

				if not self[FIELD_FROM] then return end
				temp[2]         = self[FIELD_FROM]

				if self[FIELD_WHERE] then
					temp[3]     = "WHERE"
					temp[4]     = self[FIELD_WHERE]
				else
					return
				end
			elseif sqltype == SQLTYPE_INSERT then
				temp[1]         = "INSERT INTO"

				if not self[FIELD_FROM] then return end
				temp[2]         = self[FIELD_FROM]

				if not self[FIELD_SELECT] then return end

				temp[3]         = "("
				temp[4]         = self[FIELD_SELECT]
				temp[5]         = ") VALUES ("

				if not self[FIELD_INSERT] then return end

				temp[6]         = self[FIELD_INSERT]
				temp[7]         = ")"
			end

			return tblconcat(temp, " ")
		end

		-----------------------------------------------------------
		--                      constructor                      --
		-----------------------------------------------------------
		function __new(self)
			return {
				[FIELD_SQLTYPE] = false,
				[FIELD_SELECT]  = false,
				[FIELD_UPDATE]  = false,
				[FIELD_INSERT]  = false,
				[FIELD_FROM]    = false,
				[FIELD_WHERE]   = false,
				[FIELD_ORDERBY] = false,
			}, true
		end
	end)

	__Sealed__() class "MySQLConnection" (function(_ENV)
		extend "IDbConnection"

		property "SqlBuilder" { set = false, default = MySQLBuilder }

		-----------------------------------------------------------
		--                        method                         --
		-----------------------------------------------------------
		--- Begins a database transaction.
		__Arguments__{ TransactionIsolation/TransactionIsolation.REPEATABLE_READ }
		function NewTransaction(self, isolation)
			return MySQLTransaction(self, isolation)
		end

		--- Closes the connection to the database.
		function Close(self)
			print("[Connection][Close]")
		end

		--- Opens a database connection with the settings specified by the ConnectionString property of the provider-specific Connection object.
		function Open(self)
			print("[Connection][Open]")
		end

		--- Sends the query sql to the remote MySQL server
		function Query(self, sql)
			print("[SQL][Query]", sql)
		end

		--- Sends the insert sql to the remote MySQL server
		function Insert(self, sql)
			print("[SQL][Insert]", sql)
		end

		--- Sends the update sql to the remote MySQL server
		function Update(self, sql)
			print("[SQL][Update]", sql)
		end

		function Delete(self, sql)
			print("[SQL][Delete]", sql)
		end

		function Execute(self, sql)
			print("[SQL][Execute]", sql)
		end
	end)

	__DataContext__()
	class "TestDBContext" (function(_ENV)
		-- Init the connection when a context object is created
		function __ctor(self)
			self.Connection = MySQLConnection()
		end

		__DataTable__ {
			name         = "employee",

			indexes      = {
				{ fields = { "id" },   primary = true },
				{ fields = { "name" },  unique = true },
			}
		}
		class "Employee" (function(_ENV)
			-- autoincr means the id is auto given by the data base when inserted
			__DataField__{ autoincr = true }
			property "id"           { type = NaturalNumber }

			__DataField__{ notnull = true }
			property "name"         { type = String }
		end)
	end)

	-- [Connection][Open]
	with(TestDBContext())(function(ctx)
		-- [SQL][Query] SELECT depid,name,joindate,id,leavedate,telno,email FROM employee WHERE id=1
		local e = ctx.Employees:Query{ id = 1 }:First()

		-- [SQL][Execute]   TRANSACTION BEGIN
		with(ctx.Transaction)(function(tran)
			e = ctx.Employees:Add{
				name = "Ann"
			}

			-- [SQL][Insert]    INSERT INTO employee ( name ) VALUES ( "Ann" )
			ctx:SaveChanges()
		end)
		-- [SQL][Execute]  TRANSACTION COMMIT
	end)
	-- [Connection][Close]
end)
```


## System.Data.ICache

We won't always retrieve data entities from the databases, to reduce the cost, we may save/load the entities in caches.

The caches could be a Lua table, a redis server and etc. Their usages are very different, the **ICache** interface has declared several methods to be implmented by them so we can use them for unified style usage.

Here is the definiton of the **ICache**:

```lua
--- The basic cache interface
interface "System.Data.ICache" (function(_ENV)
	extend "System.IAutoClose"

	-----------------------------------------------------------
	--                        method                         --
	-----------------------------------------------------------
	--- Set key-value pair to the cache
	__Abstract__() function Set(self, key, value, expiretime) end

	--- Set the expire time for a key
	__Abstract__() function SetExpireTime(self, key, expiretime) end

	--- Get value for a key
	__Abstract__() function Get(self, key) end

	--- Whether the key existed in the cache
	__Abstract__() function Exist(self, key) end

	--- Delete a key from the cache
	__Abstract__() function Delete(self, key) end
end)
```

Beware, since the system won't know how the cache save the entities(or any value), the serialization/deserialization must be done by the cache object.

You can find two implementation **Redis** and **ShareDict** in the [NgxLua][].

Take the example of the **Redis**:

```lua
require "NgxLua"

PLoop(function(_ENV)
	-- Open default server: 127.0.0.1:6379
	with(NgxLua.Redis())(function(redis)
		-- Add a user entity for 10 mins
		redis:Set("user_100", { id = 100, name = "Ann" }, Date.Now:AddMinutes(10))
	end)
end)
```


## System.Data.DataEntityCache

To combie the usage of the data entity and the cache system, we could use the **DataEntityCache**, this is a template class, we could pass the data entity class as its argument, the **DataEntityCache** would get the entity class's context class by get the parent namespace of it.

Abstract Property         |Type                             |Description
:-------------------------|:--------------------------------|:----------------------------
DataContext               |IDataContext                     |Gets or sets the data context, by default the entity class's parent namespace will be used to generate the context object
Cache                     |ICache                           |Gets or sets the cache object, must be done in the contstructor
Timeout                   |NaturalNumber                    |Gets or sets the timeout in seconds, default nil, no timeout


Method                    |Arguments                        |Description
:-------------------------|:--------------------------------|:----------------------------
GetEntity                 |... : Any * 0                    |Get the entity by primary keys with order
GetEntity                 |query: Table                     |Get the entity with query table like `{id = 1 }`
SaveEntity                |entity                           |Save the entity to the cache
DeleteEntity              |entity                           |Delete the entity from the cache


The **GetEntity** will try to retrieve the entity from the cache, if not exist, it will retrieve the entity from the data context and save the entity to the cache for next query.

Here is an example combined with the previous examples:

```lua
class "EmployeeCache" (function(_ENV)
	inherit (System.Data.DataEntityCache[TestDBContext.Employee])

	function __ctor(self)
		self.Cache = NgxLua.Redis()
	end
end)


function getEmployee(id)
	return with(EmployeeCache())(function(cache)
		return cache:GetEntity(id)
	end)
end
```

The **SaveEntity** method only save the entity to the cache, it won't do anything to the database, so if you update/delete the entity by the data context, you must use the **DeleteEntity** method to remove it from the cache or save a new entity to the cache.

```lua
function fireEmployee(id)
	if not id then return end

	with(TestDBContext())(function(ctx)
		-- Open the database connection

		with(ctx.Transaction)(function(trans)
			-- Start a transaction

			-- Query and Lock the employee
			local employee = ctx.employees:Lock{ id = id }:First()

			if employee the
				employee.leavedate = Date.Now

				-- Save the operation to database
				ctx:SaveChanges()

				with(EmployeeCache())(function(cache)
					cache:SaveEntity(employee)
				end)
			else
				-- Cancel the transaction
				trans:Rollback()
			end
		end)
	end)
end
```


[NgxLua]: https://github.com/kurapica/NgxLua/ "The implementation for Openresty"