# Multi Os Thread Platform

It's very different to use Lua on a single os thread platforms(in most case) and the multi os thread platforms(like the [Openresty][]), there are many import tips to help you avoid the conflict between threads.


## use export

The environment in **PLoop** will auto cache the variables the code accessed during the run-time, but because the re-hash of the environment isn't thread safe, the environment has to use a proxy table to save those variables, the proxy is always replaced by a new one when new variables is added, the cost is high, and assign new variables to the environment is also dangerous.

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	function _G.test()
		local v = math.random(100)
	end
end)

-- [PLoop: Warn]The [math] is auto saved to table: 010AE1D8, need use 'export{ "math" }'@xxxxx.lua:5	4
test()
```

With the **MULTI_OS_THREAD** setting, the **PLoop** will *Warn* the usages of auto cache behaviors, it's best to follow the message, use the **export** to solve the problem:

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	export { "math" }
	function _G.test()
		local v = math.random(100)
	end
end)
```

The **export** keyword will save those variables into the environment during the definition time, you can also assign the variables to false value if you need it later.


## use System.Toolset.safeset

If you need use a table to cache datas, and the table may be accessed through several threads, it's better to use the **safeset** method, it'll clone a new table used to replaced the old one.

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	export { safeset = Toolset.safeset }

	local cache = {}

	function add(name, value)
		cache = safeset(cache, name, value)
	end
end)
```

Beware, this have no use for operation codes like `cache.x = cache.x + 1`. It's still not thread safe for operations.


## use Lock for thread-safe

The **System.Threading** has provided a **ILockManager** interface, you may find an implementation **NgxLua.LockManager** in [NgxLua][], there should be only one manager existed in a system.

The interface is defined :

Static Property      |Description
:--------------------|:-----------------------------
Manager              |The manager object


Static Method        |Arguments                      |Description
:--------------------|:------------------------------|:-------------------------
RunWithLock          |key: Any, func: Function, ...  |Lock with a key and process the target function, return the values
TryRunWithLock       |key: Any, func: Function, ...  |Try lock with a key and process the target function, return the values


Abstract Method      |Arguments                      |Description
:--------------------|:------------------------------|:-------------------------
Lock                 |key: Any                       |Lock with a key and return a lock object to release
TryLock              |key: Any                       |Try lock with a key and return a lock object to release
Release              |obj: Any, key: Any             |Release the lock object

We also can use the attribute `__Lock__` and `__TryLock__` instead of call the **RunWithLock** directly.

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	local cache = {}

	__Lock__("TEST_ADD_KEY")
	function add(name, value)
		cache[name] = value
	end
end)
```


## System.Context

In the system like [Openresty][], each os thread processed a coroutine to do the tasks. Besides share the data between the coroutines in different os threads, share data in the coroutine is also a problem, we can't use a table to map the coroutine to table, the root table also will cause problems.

With the **debug.getlocal** api, the **PLoop** has provided the **Context** system. the **System.Web.HttpContext** is also inherit from it.

The **Context** should be the root object that start the task process in a os thread. So all object generated in it will share the same context object, they can saved any variables in the context and use them.

Here is its features:

Static Property   |Description
:-----------------|:-------------------------
Current           |Gets the current context object


Static Method         |Arguments          |Description
:---------------------|:------------------|:-------------
GetContextFromStack   |stack: Number/nil  |Retrieve the context object from the begin stack to the top stack


Abstract Method   |Description
:-----------------|:---------------------------
Process           |Process the operations under the context


If any class object need use the **Context** object, it'd extend the **System.IContext** interface, it provide a final property:

Final Property    |Description
:-----------------|:---------------------------
Context           |The context where the object is created


Here is an example:

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	class "A" { IContext }

	ctx = Context()

	function getObj()
		return A()
	end

	function ctx:Process()
		obj = getObj()

		print(obj.Context == self)  -- true
	end

	ctx:Process()
end)
```

To save data in the context, we must choose special key to avoid conflict, normally use the class is a good choice.

The **ThreadPool** will also works with the context system.

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	__Async__() function test()
		print(coroutine.running())
	end

	test()  -- thread: 00E8A988
	test()  -- thread: 00E8B388
end)
```

The coroutines won't be recycled in the multi os thread mode, since it's not safe to share the default pool.

The coroutines are heavily used in the **PLoop**, we should use standalone thread pool for each os thread:

```lua
PLOOP_PLATFORM_SETTINGS = { MULTI_OS_THREAD = true }

require "PLoop" (function(_ENV)
	__Async__() function test()
		print(coroutine.running())
	end

	ctx = Context()

	function ctx:Process()
		test()  -- thread: 00DD2578
		test()  -- thread: 00DD2578
	end

	ctx:Process()
end)
```

So in a context, we'll have a standalone thread pool to recycle the coroutines.


[Openresty]: https://github.com/openresty/lua-nginx-module/ "Openresty"
[NgxLua]: https://github.com/kurapica/NgxLua/ "An implementation for the Openresty"